#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, division, print_function

from builtins import str
from builtins import object

profiling = False  # turning on will save profile files in currDir

import sys
import argparse
import platform
import psychopy
from psychopy import prefs
from pkg_resources import parse_version
from psychopy.constants import PY3
from . import urls
from . import frametracker
from . import themes
from . import icons

import io
import json

if not hasattr(sys, 'frozen'):
    try:
        import wxversion
        haveWxVersion = True
    except ImportError:
        haveWxVersion = False  # if wxversion doesn't exist hope for the best
    if haveWxVersion:
        wxversion.ensureMinimal('2.8')  # because this version has agw
import wx
try:
    from agw import advancedsplash as AS
except ImportError:  # if it's not there locally, try the wxPython lib.
    import wx.lib.agw.advancedsplash as AS

# from .plugin_manager import saveStartUpPluginsConfig

from psychopy.localization import _translate
# NB keep imports to a minimum here because splash screen has not yet shown
# e.g. coder and builder are imported during app.__init__ because they
# take a while

# needed by splash screen for the path to resources/psychopySplash.png
import ctypes
from psychopy import logging, __version__
from psychopy import projects
from . import connections
from .utils import FileDropTarget
import os
import weakref

# knowing if the user has admin priv is generally a good idea for security.
# not actually needed; psychopy should never need anything except normal user
# see older versions for code to detect admin (e.g., v 1.80.00)

if not PY3 and sys.platform == 'darwin':
    blockTips = True
else:
    blockTips = False

travisCI = bool(str(os.environ.get('TRAVIS')).lower() == 'true')

# Enable high-dpi support if on Windows. This fixes blurry text rendering.
if sys.platform == 'win32':
    # get the preference for high DPI
    if 'highDPI' in psychopy.prefs.app.keys():  # check if we have the option
        enableHighDPI = psychopy.prefs.app['highDPI']

        # check if we have OS support for it
        if enableHighDPI:
            try:
                ctypes.windll.shcore.SetProcessDpiAwareness(enableHighDPI)
            except OSError:
                logging.warn(
                    "High DPI support is not appear to be supported by this version"
                    " of Windows. Disabling in preferences.")

                psychopy.prefs.app['highDPI'] = False
                psychopy.prefs.saveUserPrefs()


class MenuFrame(wx.Frame, themes.ThemeMixin):
    """A simple empty frame with a menubar, should be last frame closed on mac
    """

    def __init__(self, parent=None, ID=-1, app=None, title="PsychoPy3"):

        wx.Frame.__init__(self, parent, ID, title, size=(1, 1))
        self.app = app

        self.menuBar = wx.MenuBar()

        self.viewMenu = wx.Menu()
        self.menuBar.Append(self.viewMenu, _translate('&View'))
        mtxt = _translate("&Open Builder view\t%s")
        self.app.IDs.openBuilderView = self.viewMenu.Append(wx.ID_ANY,
                             mtxt,
                             _translate("Open a new Builder view")).GetId()
        self.Bind(wx.EVT_MENU, self.app.showBuilder,
                  id=self.app.IDs.openBuilderView)
        mtxt = _translate("&Open Coder view\t%s")
        self.app.IDs.openCoderView = self.viewMenu.Append(wx.ID_ANY,
                             mtxt,
                             _translate("Open a new Coder view")).GetId()
        self.Bind(wx.EVT_MENU, self.app.showCoder,
                  id=self.app.IDs.openCoderView)
        mtxt = _translate("&Quit\t%s")
        item = self.viewMenu.Append(wx.ID_EXIT, mtxt % self.app.keys['quit'],
                                    _translate("Terminate the program"))
        self.Bind(wx.EVT_MENU, self.app.quit, id=item.GetId())
        self.SetMenuBar(self.menuBar)
        self.Show()


class IDStore(dict):
    """A simpe class that works like a dict but you can access attributes
    like standard python attrs. Useful to replace the previous pre-made
    app.IDs (wx.NewID() is no longer recommended or safe)
    """
    def __getattr__(self, attr):
        return self[attr]
    def __setattr__(self, attr, value):
        self[attr] = value


class _Showgui_Hack(object):
    """Class with side-effect of restoring wx window switching under wx-3.0

    - might only be needed on some platforms (Mac 10.9.4 needs it for me);
    - needs to be launched as an external script
    - needs to be separate: seg-faults as method of PsychoPyApp or in-lined
    - unlear why it works or what the deeper issue is, blah
    - called at end of PsychoPyApp.onInit()
    """

    def __init__(self):
        super(_Showgui_Hack, self).__init__()
        from psychopy import core
        import os
        # should be writable:
        noopPath = os.path.join(psychopy.prefs.paths['userPrefsDir'],
                                'showgui_hack.py')
        # code to open & immediately close a gui (= invisibly):
        if not os.path.isfile(noopPath):
            code = """from psychopy import gui
                dlg = gui.Dlg().Show()  # non-blocking
                try: 
                    dlg.Destroy()  # might as well
                except Exception: 
                    pass"""
            with open(noopPath, 'wb') as fd:
                fd.write(bytes(code))
        # append 'w' for pythonw seems not needed
        core.shellCall([sys.executable, noopPath])


class PsychoPyApp(wx.App, themes.ThemeMixin):

    def __init__(self, arg=0, testMode=False, **kwargs):
        """With a wx.App some things get done here, before App.__init__
        then some further code is launched in OnInit() which occurs after
        """
        if profiling:
            import cProfile, time
            profile = cProfile.Profile()
            profile.enable()
            t0 = time.time()

        self._appLoaded = False  # set to true when all frames are created
        self.coder = None
        self.runner = None
        self.version = psychopy.__version__
        # set default paths and prefs
        self.prefs = psychopy.prefs
        self._currentThemeSpec = None

        self.keys = self.prefs.keys
        self.prefs.pageCurrent = 0  # track last-viewed page, can return there
        self.IDs = IDStore()
        self.urls = urls.urls
        self.quitting = False
        # check compatibility with last run version (before opening windows)
        self.firstRun = False
        self.testMode = testMode
        self._stdout = sys.stdout
        self._stderr = sys.stderr
        self._stdoutFrame = None
        self.iconCache = themes.IconCache()

        if not self.testMode:
            self._lastRunLog = open(os.path.join(
                    self.prefs.paths['userPrefsDir'], 'last_app_load.log'),
                    'w')
            sys.stderr = sys.stdout = lastLoadErrs = self._lastRunLog
            logging.console.setLevel(logging.DEBUG)

        # indicates whether we're running for testing purposes
        self.osfSession = None
        self.pavloviaSession = None

        self.copiedRoutine = None
        self.copiedCompon = None
        self._allFrames = frametracker.openFrames  # ordered; order updated with self.onNewTopWindow

        wx.App.__init__(self, arg)

        # import localization after wx:
        from psychopy import localization  # needed by splash screen
        self.localization = localization
        self.locale = localization.setLocaleWX()
        self.locale.AddCatalog(self.GetAppName())

        # set the exception hook to present unhandled errors in a dialog
        if not travisCI:
            from psychopy.app.errorDlg import exceptionCallback
            sys.excepthook = exceptionCallback

        self.onInit(testMode=testMode, **kwargs)
        if profiling:
            profile.disable()
            print("time to load app = {:.2f}".format(time.time()-t0))
            profile.dump_stats('profileLaunchApp.profile')


    def onInit(self, showSplash=True, testMode=False):
        """This is launched immediately *after* the app initialises with wx
        :Parameters:

          testMode: bool
        """
        self.SetAppName('PsychoPy3')

        if showSplash: #showSplash:
            # show splash screen
            splashFile = os.path.join(
                self.prefs.paths['resources'], 'psychopySplash.png')
            splashImage = wx.Image(name=splashFile)
            splashImage.ConvertAlphaToMask()
            splash = AS.AdvancedSplash(None, bitmap=splashImage.ConvertToBitmap(),
                                       timeout=3000,
                                       agwStyle=AS.AS_TIMEOUT | AS.AS_CENTER_ON_SCREEN,
                                       )  # transparency?
            w, h = splashImage.GetSize()
            splash.SetTextPosition((int(340), h-30))
            splash.SetText(_translate("Copyright (C) 2020 OpenScienceTools.org"))
        else:
            splash = None

        # SLOW IMPORTS - these need to be imported after splash screen starts
        # but then that they end up being local so keep track in self

        from psychopy.compatibility import checkCompatibility
        # import coder and builder here but only use them later
        from psychopy.app import coder, builder, runner, dialogs

        if '--firstrun' in sys.argv:
            del sys.argv[sys.argv.index('--firstrun')]
            self.firstRun = True
        if 'lastVersion' not in self.prefs.appData:
            # must be before 1.74.00
            last = self.prefs.appData['lastVersion'] = '1.73.04'
            self.firstRun = True
        else:
            last = self.prefs.appData['lastVersion']

        if self.firstRun and not self.testMode:
            pass

        # setup links for URLs
        # on a mac, don't exit when the last frame is deleted, just show menu
        if sys.platform == 'darwin':
            self.menuFrame = MenuFrame(parent=None, app=self)
        # fetch prev files if that's the preference
        if self.prefs.coder['reloadPrevFiles']:
            scripts = self.prefs.appData['coder']['prevFiles']
        else:
            scripts = []
        appKeys = list(self.prefs.appData['builder'].keys())
        if self.prefs.builder['reloadPrevExp'] and ('prevFiles' in appKeys):
            exps = self.prefs.appData['builder']['prevFiles']
        else:
            exps = []
        runlist = []

        self.dpi = int(wx.GetDisplaySize()[0] /
                       float(wx.GetDisplaySizeMM()[0]) * 25.4)
        if not (50 < self.dpi < 120):
            self.dpi = 80  # dpi was unreasonable, make one up

        if sys.platform == 'win32':
            # wx.SYS_DEFAULT_GUI_FONT is default GUI font in Win32
            self._mainFont = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        else:
            self._mainFont = wx.SystemSettings.GetFont(wx.SYS_ANSI_FIXED_FONT)

        try:
            self._codeFont = wx.SystemSettings.GetFont(wx.SYS_ANSI_FIXED_FONT)
        except wx._core.wxAssertionError:
            # if no SYS_ANSI_FIXED_FONT then try generic FONTFAMILY_MODERN
            self._codeFont = wx.Font(self._mainFont.GetPointSize(),
                                     wx.FONTFAMILY_MODERN,
                                     wx.FONTSTYLE_NORMAL,
                                     wx.FONTWEIGHT_NORMAL)
        # that gets most of the properties of _codeFont but the FaceName
        # FaceName is set in the setting of the theme:
        self.theme = self.prefs.app['theme']

        # removed Aug 2017: on newer versions of wx (at least on mac)
        # this looks too big
        # if hasattr(self._mainFont, 'Larger'):
        #     # Font.Larger is available since wyPython version 2.9.1
        #     # PsychoPy still supports 2.8 (see ensureMinimal above)
        #     self._mainFont = self._mainFont.Larger()
        #     self._codeFont.SetPointSize(
        #         self._mainFont.GetPointSize())  # unify font size

        # create both frame for coder/builder as necess
        if splash:
            splash.SetText(_translate("  Creating frames..."))

        # Parse incoming call
        parser = argparse.ArgumentParser(prog=self)
        parser.add_argument('--builder', dest='builder', action="store_true")
        parser.add_argument('-b', dest='builder', action="store_true")
        parser.add_argument('--coder', dest='coder', action="store_true")
        parser.add_argument('-c', dest='coder', action="store_true")
        parser.add_argument('--runner', dest='runner', action="store_true")
        parser.add_argument('-r', dest='runner', action="store_true")
        view, args = parser.parse_known_args(sys.argv)
        print(args)
        # Check from filetype if any windows need to be open
        if any(arg.endswith('.psyexp') for arg in args):
            view.builder = True
            exps = [file for file in args if file.endswith('.psyexp')]
        if any(arg.endswith('.psyrun') for arg in args):
            view.runner = True
            runlist = [file for file in args if file.endswith('.psyrun')]
        # If still no window specified, use default from prefs
        if not any(getattr(view, key) for key in ['builder', 'coder', 'runner']):
            if self.prefs.app['defaultView'] in view:
                setattr(view, self.prefs.app['defaultView'], True)
            elif self.prefs.app['defaultView'] == 'all':
                view.builder = True
                view.coder = True
                view.runner = True

        # Create windows
        if view.runner:
            self.showRunner(fileList=runlist)
        if view.coder:
            self.showCoder(fileList=scripts)
        if view.builder:
            self.showBuilder(fileList=exps)

        # send anonymous info to www.psychopy.org/usage.php
        # please don't disable this, it's important for PsychoPy's development
        self._latestAvailableVersion = None
        self.updater = None
        self.news = None
        self.tasks = None

        prefsConn = self.prefs.connections

        ok, msg = checkCompatibility(last, self.version, self.prefs, fix=True)
        # tell the user what has changed
        if not ok and not self.firstRun and not self.testMode:
            title = _translate("Compatibility information")
            dlg = dialogs.MessageDialog(parent=None, message=msg, type='Info',
                                        title=title)
            dlg.ShowModal()

        if (self.prefs.app['showStartupTips']
                and not self.testMode and not blockTips):
            tipFile = os.path.join(
                self.prefs.paths['resources'], _translate("tips.txt"))
            tipIndex = self.prefs.appData['tipIndex']
            if parse_version(wx.__version__) >= parse_version('4.0.0a1'):
                tp = wx.adv.CreateFileTipProvider(tipFile, tipIndex)
                showTip = wx.adv.ShowTip(None, tp)
            else:
                tp = wx.CreateFileTipProvider(tipFile, tipIndex)
                showTip = wx.ShowTip(None, tp)

            self.prefs.appData['tipIndex'] = tp.GetCurrentTip()
            self.prefs.saveAppData()
            self.prefs.app['showStartupTips'] = showTip
            self.prefs.saveUserPrefs()

        self.Bind(wx.EVT_IDLE, self.onIdle)

        # doing this once subsequently enables the app to open & switch among
        # wx-windows on some platforms (Mac 10.9.4) with wx-3.0:
        v = parse_version
        if sys.platform == 'darwin':
            if v('3.0') <= v(wx.version()) <v('4.0'):
                _Showgui_Hack()  # returns ~immediately, no display
                # focus stays in never-land, so bring back to the app:
                if prefs.app['defaultView'] in ['all', 'builder', 'coder', 'runner']:
                    self.showBuilder()
                else:
                    self.showCoder()
        # after all windows are created (so errors flushed) create output
        self._appLoaded = True
        if self.coder:
            self.coder.setOutputWindow()  # takes control of sys.stdout

        # flush any errors to the last run log file
        logging.flush()
        sys.stdout.flush()
        # we wanted debug mode while loading but safe to go back to info mode
        if not self.prefs.app['debugMode']:
            logging.console.setLevel(logging.INFO)

        return True

    def _wizard(self, selector, arg=''):
        from psychopy import core
        wizard = os.path.join(
            self.prefs.paths['psychopy'], 'tools', 'wizard.py')
        so, se = core.shellCall(
            [sys.executable, wizard, selector, arg], stderr=True)
        if se and self.prefs.app['debugMode']:
            print(se)  # stderr contents; sometimes meaningless

    def firstrunWizard(self):
        self._wizard('--config', '--firstrun')
        # wizard typically creates html report file but user can manually skip
        reportPath = os.path.join(
            self.prefs.paths['userPrefsDir'], 'firstrunReport.html')
        if os.path.exists(reportPath):
            with io.open(reportPath, 'r', encoding='utf-8-sig') as f:
                report = f.read()
            if 'Configuration problem' in report:
                # fatal error was encountered (currently only if bad drivers)
                # ensure wizard will be triggered again:
                del self.prefs.appData['lastVersion']
                self.prefs.saveAppData()

    def benchmarkWizard(self, evt=None):
        self._wizard('--benchmark')

    def csvFromPsydat(self, evt=None):
        from psychopy import gui
        from psychopy.tools.filetools import fromFile

        prompt = _translate("Select .psydat file(s) to extract")
        names = gui.fileOpenDlg(allowed='*.psydat', prompt=prompt)
        for name in names or []:
            filePsydat = os.path.abspath(name)
            print("psydat: {0}".format(filePsydat))

            exp = fromFile(filePsydat)
            if filePsydat.endswith('.psydat'):
                fileCsv = filePsydat[:-7]
            else:
                fileCsv = filePsydat
            fileCsv += '.csv'
            exp.saveAsWideText(fileCsv)
            print('   -->: {0}'.format(os.path.abspath(fileCsv)))

    def checkUpdates(self, evt):
        # if we have internet and haven't yet checked for updates then do so
        # we have a network connection but not yet tried an update
        if self._latestAvailableVersion not in [-1, None]:
            # change IDLE routine so we won't come back here
            self.Unbind(wx.EVT_IDLE)  # unbind all EVT_IDLE methods from app
            self.Bind(wx.EVT_IDLE, self.onIdle)
            # create updater (which will create dialogs as needed)
            self.updater = connections.Updater(app=self)
            self.updater.latest = self._latestAvailableVersion
            self.updater.suggestUpdate(confirmationDlg=False)
        evt.Skip()

    def getPrimaryDisplaySize(self):
        """Get the size of the primary display (whose coords start (0,0))
        """
        return list(wx.Display(0).GetGeometry())[2:]

    def makeAccelTable(self):
        """Makes a standard accelorator table and returns it. This then needs
        to be set for the Frame using self.SetAccelerator(table)
        """
        def parseStr(inStr):
            accel = 0
            if 'ctrl' in inStr.lower():
                accel += wx.ACCEL_CTRL
            if 'shift' in inStr.lower():
                accel += wx.ACCEL_SHIFT
            if 'alt' in inStr.lower():
                accel += wx.ACCEL_ALT
            return accel, ord(inStr[-1])
        # create a list to link IDs to key strings
        keyCodesDict = {}
        keyCodesDict[self.keys['copy']] = wx.ID_COPY
        keyCodesDict[self.keys['cut']] = wx.ID_CUT
        keyCodesDict[self.keys['paste']] = wx.ID_PASTE
        keyCodesDict[self.keys['undo']] = wx.ID_UNDO
        keyCodesDict[self.keys['redo']] = wx.ID_REDO
        keyCodesDict[self.keys['save']] = wx.ID_SAVE
        keyCodesDict[self.keys['saveAs']] = wx.ID_SAVEAS
        keyCodesDict[self.keys['close']] = wx.ID_CLOSE
        keyCodesDict[self.keys['redo']] = wx.ID_REDO
        keyCodesDict[self.keys['quit']] = wx.ID_EXIT
        # parse the key strings and convert to accelerator entries
        entries = []
        for keyStr, code in list(keyCodesDict.items()):
            mods, key = parseStr(keyStr)
            entry = wx.AcceleratorEntry(mods, key, code)
            entries.append(entry)
        table = wx.AcceleratorTable(entries)
        return table

    def updateWindowMenu(self):
        """Update items within Window menu to reflect open windows"""
        # Update checks on menus in all frames
        for frame in self.getAllFrames():
            if hasattr(frame, "windowMenu"):
                frame.windowMenu.updateFrames()

    def showCoder(self, event=None, fileList=None):
        # have to reimport because it is only local to __init__ so far
        from . import coder
        if self.coder is None:
            title = "PsychoPy3 Coder (IDE) (v%s)"
            wx.BeginBusyCursor()
            self.coder = coder.CoderFrame(None, -1,
                                          title=title % self.version,
                                          files=fileList, app=self)
            self.updateWindowMenu()
            wx.EndBusyCursor()
        else:
            # Set output window and standard streams
            self.coder.setOutputWindow(True)
        self.coder.Show(True)
        self.SetTopWindow(self.coder)
        self.coder.Raise()

    def newBuilderFrame(self, event=None, fileName=None):
        # have to reimport because it is ony local to __init__ so far
        wx.BeginBusyCursor()
        from .builder.builder import BuilderFrame
        title = "PsychoPy3 Experiment Builder (v%s)"
        self.builder = BuilderFrame(None, -1,
                                 title=title % self.version,
                                 fileName=fileName, app=self)
        self.builder.Show(True)
        self.builder.Raise()
        self.SetTopWindow(self.builder)
        self.updateWindowMenu()
        wx.EndBusyCursor()
        return self.builder

    def showBuilder(self, event=None, fileList=()):
        # have to reimport because it is only local to __init__ so far
        from psychopy.app import builder
        for fileName in fileList:
            if os.path.isfile(fileName):
                self.newBuilderFrame(fileName=fileName)
        # create an empty Builder view if needed
        if len(self.getAllFrames(frameType="builder")) == 0:
            self.newBuilderFrame()
        # loop through all frames, from the back bringing each forward
        for thisFrame in self.getAllFrames(frameType='builder'):
            thisFrame.Show(True)
            thisFrame.Raise()
            self.SetTopWindow(thisFrame)

    def showRunner(self, event=None, fileList=[]):
        if not self.runner:
            self.runner = self.newRunnerFrame()
        if not self.testMode:
            self.runner.Show()
            self.runner.Raise()
            self.SetTopWindow(self.runner)
        # Runner captures standard streams until program closed
        if self.runner and not self.testMode:
            sys.stdout = self.runner.stdOut
            sys.stderr = self.runner.stdOut

    def newRunnerFrame(self, event=None):
        # have to reimport because it is only local to __init__ so far
        from .runner.runner import RunnerFrame
        title = "PsychoPy3 Experiment Runner (v{})".format(self.version)
        wx.BeginBusyCursor()
        self.runner = RunnerFrame(parent=None,
                             id=-1,
                             title=title,
                             app=self)
        self.updateWindowMenu()
        wx.EndBusyCursor()
        return self.runner

    def OnDrop(self, x, y, files):
        """Not clear this method ever gets called!"""
        logging.info("Got Files")

    def MacOpenFile(self, fileName):
        if fileName.endswith('psychopyApp.py'):
            # in wx4 on mac this is called erroneously by App.__init__
            # if called like `python psychopyApp.py`
            return
        logging.debug('PsychoPyApp: Received Mac file dropped event')
        if fileName.endswith('.py'):
            if self.coder is None:
                self.showCoder()
            self.coder.setCurrentDoc(fileName)
        elif fileName.endswith('.psyexp'):
            self.newBuilderFrame(fileName=fileName)

    def MacReopenApp(self):
        """Called when the doc icon is clicked, and ???"""
        self.GetTopWindow().Raise()

    def openIPythonNotebook(self, event=None):
        """Note that right now this is bad because it ceases all activity in
        the main wx loop and the app has to be quit. We need it to run from
        a separate process? The necessary depends (zmq and tornado) were
        included from v1.78 onwards in the standalone
        """
        import IPython.frontend.html.notebook.notebookapp as nb
        instance = nb.launch_new_instance()

    def openUpdater(self, event=None):
        from psychopy.app import connections
        dlg = connections.InstallUpdateDialog(parent=None, ID=-1, app=self)

    def colorPicker(self, event=None):
        """Open color-picker, sets clip-board to string [r,g,b].

        Note: units are psychopy -1..+1 rgb units to three decimal places,
        preserving 24-bit color.
        """
        class ColorPicker(wx.Panel):

            def __init__(self, parent):
                wx.Panel.__init__(self, parent, wx.ID_ANY)
                rgb = 'None'
                dlg = wx.ColourDialog(self)
                dlg.GetColourData().SetChooseFull(True)
                if dlg.ShowModal() == wx.ID_OK:
                    data = dlg.GetColourData()
                    rgb = data.GetColour().Get(includeAlpha=False)
                    rgb = map(lambda x: "%.3f" %
                              ((x - 127.5) / 127.5), list(rgb))
                    rgb = '[' + ','.join(rgb) + ']'
                    # http://wiki.wxpython.org/AnotherTutorial#wx.TheClipboard
                    if wx.TheClipboard.Open():
                        wx.TheClipboard.Clear()
                        wx.TheClipboard.SetData(wx.TextDataObject(str(rgb)))
                        wx.TheClipboard.Close()
                dlg.Destroy()
                parent.newRBG = rgb
        frame = wx.Frame(None, wx.ID_ANY, "Color picker",
                         size=(0, 0))  # not shown
        ColorPicker(frame)
        newRBG = frame.newRBG
        frame.Destroy()
        return newRBG  # string

    def openMonitorCenter(self, event):
        from psychopy.monitors import MonitorCenter
        self.monCenter = MonitorCenter.MainFrame(
            None, 'PsychoPy3 Monitor Center')
        self.monCenter.Show(True)

    def terminateHubProcess(self):
        """
        Send a UDP message to iohub informing it to exit.

        Use this when force quitting the experiment script process so iohub
        knows to exit as well.

        If message is not sent within 1 second, or the iohub server
        address in incorrect,the issue is logged.
        """
        sock = None
        try:
            logging.debug('PsychoPyApp: terminateHubProcess called.')
            import socket
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.settimeout(1.0)
            iohubAddress = '127.0.0.1', 9034
            import msgpack
            txData = msgpack.Packer().pack(('STOP_IOHUB_SERVER',))
            return sock.sendto(txData, iohubAddress)
        except socket.error as e:
            msg = 'PsychoPyApp: terminateHubProcess socket.error: %s'
            logging.debug(msg % str(e))
        except socket.herror as e:
            msg = 'PsychoPyApp: terminateHubProcess socket.herror: %s'
            logging.debug(msg % str(e))
        except socket.gaierror as e:
            msg = 'PsychoPyApp: terminateHubProcess socket.gaierror: %s'
            logging.debug(msg % str(e))
        except socket.timeout as e:
            msg = 'PsychoPyApp: terminateHubProcess socket.timeout: %s'
            logging.debug(msg % str(e))
        except Exception as e:
            msg = 'PsychoPyApp: terminateHubProcess exception: %s'
            logging.debug(msg % str(e))
        finally:
            if sock:
                sock.close()
            logging.debug('PsychoPyApp: terminateHubProcess completed.')

    def quit(self, event=None):
        logging.debug('PsychoPyApp: Quitting...')
        self.quitting = True
        # garbage collect the projects before sys.exit
        projects.pavlovia.knownUsers = None
        projects.pavlovia.knownProjects = None
        # see whether any files need saving
        for frame in self.getAllFrames():
            try:  # will fail if the frame has been shut somehow elsewhere
                ok = frame.checkSave()
            except Exception:
                ok = False
                logging.debug("PsychopyApp: exception when saving")
            if not ok:
                logging.debug('PsychoPyApp: User cancelled shutdown')
                return  # user cancelled quit

        # save info about current frames for next run
        if self.coder and len(self.getAllFrames("builder")) == 0:
            self.prefs.appData['lastFrame'] = 'coder'
        elif self.coder is None:
            self.prefs.appData['lastFrame'] = 'builder'
        else:
            self.prefs.appData['lastFrame'] = 'both'

        self.prefs.appData['lastVersion'] = self.version
        # update app data while closing each frame
        # start with an empty list to be appended by each frame
        self.prefs.appData['builder']['prevFiles'] = []
        self.prefs.appData['coder']['prevFiles'] = []

        # write plugins config if changed during the session
        # saveStartUpPluginsConfig()

        for frame in self.getAllFrames():
            try:
                frame.closeFrame(event=event, checkSave=False)
                # must do this before destroying the frame?
                self.prefs.saveAppData()
            except Exception:
                pass  # we don't care if this fails - we're quitting anyway
        self.Destroy()

        # Reset streams back to default
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

        if not self.testMode:
            sys.exit()

    def showPrefs(self, event):
        from psychopy.app.preferencesDlg import PreferencesDlg
        logging.debug('PsychoPyApp: Showing prefs dlg')
        prefsDlg = PreferencesDlg(app=self)
        prefsDlg.ShowModal()
        prefsDlg.Destroy()

    def showAbout(self, event):
        logging.debug('PsychoPyApp: Showing about dlg')

        with io.open(os.path.join(self.prefs.paths['psychopy'],'LICENSE.txt'),
                     'r', encoding='utf-8-sig') as f:
            license = f.read()

        msg = _translate(
            "For stimulus generation and experimental control in python.\n"
            "PsychoPy depends on your feedback. If something doesn't work\n"
            "then let us know at psychopy-users@googlegroups.com")
        if parse_version(wx.__version__) >= parse_version('4.0a1'):
            info = wx.adv.AboutDialogInfo()
            showAbout = wx.adv.AboutBox
        else:
            info = wx.AboutDialogInfo()
            showAbout = wx.AboutBox
        if wx.version() >= '3.':
            icon = os.path.join(self.prefs.paths['resources'], 'psychopy.png')
            info.SetIcon(wx.Icon(icon, wx.BITMAP_TYPE_PNG, 128, 128))
        info.SetName('PsychoPy')
        info.SetVersion('v' + psychopy.__version__)
        info.SetDescription(msg)

        info.SetCopyright('(C) 2002-2018 Jonathan Peirce')
        info.SetWebSite('http://www.psychopy.org')
        info.SetLicence(license)
        info.AddDeveloper('Jonathan Peirce')
        info.AddDeveloper('Jeremy Gray')
        info.AddDeveloper('Sol Simpson')
        info.AddDeveloper(u'Jonas Lindel\xF8v')
        info.AddDeveloper('Yaroslav Halchenko')
        info.AddDeveloper('Erik Kastman')
        info.AddDeveloper('Michael MacAskill')
        info.AddDeveloper('Hiroyuki Sogo')
        info.AddDeveloper('David Bridges')
        info.AddDocWriter('Jonathan Peirce')
        info.AddDocWriter('Jeremy Gray')
        info.AddDocWriter('Rebecca Sharman')
        info.AddTranslator('Hiroyuki Sogo')

        if not self.testMode:
            showAbout(info)

    def showNews(self, event=None):
        connections.showNews(self, checkPrev=False)

    def showSystemInfo(self, event=None):
        """Show system information."""
        from psychopy.app.sysInfoDlg import SystemInfoDialog
        dlg = SystemInfoDialog(None)
        dlg.Show()

    def followLink(self, event=None, url=None):
        """Follow either an event id (= a key to an url defined in urls.py)
        or follow a complete url (a string beginning "http://")
        """
        if event is not None:
            wx.LaunchDefaultBrowser(self.urls[event.GetId()])
        elif url is not None:
            wx.LaunchDefaultBrowser(url)

    def getAllFrames(self, frameType=None):
        """Get a list of frames, optionally filtered by a particular kind
        (which can be "builder", "coder", "project")
        """
        frames = []
        for frameRef in self._allFrames:
            frame = frameRef()
            if (not frame):
                self._allFrames.remove(frameRef)  # has been deleted
                continue
            elif frameType and frame.frameType != frameType:
                continue
            frames.append(frame)
        return frames

    def trackFrame(self, frame):
        """Keep track of an open frame (stores a weak reference to the frame
        which will probably have a regular reference to the app)
        """
        self._allFrames.append(weakref.ref(frame))

    def forgetFrame(self, frame):
        """Keep track of an open frame (stores a weak reference to the frame
        which will probably have a regular reference to the app)
        """
        for entry in self._allFrames:
            if entry() == frame:  # is a weakref
                self._allFrames.remove(entry)

    def onIdle(self, evt):
        from . import idle
        idle.doIdleTasks(app=self)
        evt.Skip()

    def onThemeChange(self, event):
        """Handles a theme change event (from a window with a themesMenu)"""
        win = event.EventObject.Window
        newTheme = win.themesMenu.FindItemById(event.GetId()).ItemLabel
        prefs.app['theme'] = newTheme
        prefs.saveUserPrefs()
        self.theme = newTheme

    @property
    def theme(self):
        """The theme to be used through the application"""
        return prefs.app['theme']

    @theme.setter
    def theme(self, value):
        """The theme to be used through the application"""
        themes.ThemeMixin.loadThemeSpec(self, themeName=value)
        prefs.app['theme'] = value
        self._currentThemeSpec = themes.ThemeMixin.spec
        codeFont = themes.ThemeMixin.codeColors['base']['font']

        # On OSX 10.15 Catalina at least calling SetFaceName with 'AppleSystemUIFont' fails.
        # So this fix checks to see if changing the font name invalidates the font.
        # if so rollback to the font before attempted change.
        # Note that wx.Font uses referencing and copy-on-write so we need to force creation of a copy
        # witht he wx.Font() call. Otherwise you just get reference to the font that gets borked by SetFaceName()
        # -Justin Ales
        beforesetface = wx.Font(self._codeFont)
        success = self._codeFont.SetFaceName(codeFont)
        if not (success):
            self._codeFont = beforesetface
        # Apply theme
        self._applyAppTheme()

    def _applyAppTheme(self):
        """Overrides ThemeMixin for this class"""
        self.iconCache.setTheme(themes.ThemeMixin)

        for frameRef in self._allFrames:
            frame = frameRef()
            if hasattr(frame, '_applyAppTheme'):
                frame._applyAppTheme()


if __name__ == '__main__':
    # never run; stopped earlier at cannot do relative import in a non-package
    sys.exit("Do not launch the app from this script -"
             "use python psychopyApp.py instead")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""A Backend class defines the core low-level functions required by a Window
class, such as the ability to create an OpenGL context and flip the window.

Users simply call visual.Window(..., winType='pyglet') and the winType is then
used by backends.getBackend(winType) which will locate the appropriate class
and initialize an instance using the attributes of the Window.
"""

from __future__ import absolute_import, print_function
from builtins import object
import weakref

from psychopy import logging
from psychopy.tools.attributetools import attributeSetter


class BaseBackend(object):
    """The backend class provides all the core low-level functions required by
    a Window class, such as the ability to create an OpenGL context and flip
    the window.

    Users simply call visual.Window(..., winType='pyglet') and the winType is
    then used by backends.getBackend(winType) which will locate the appropriate
    class and initialize an instance using the attributes of the Window.
    """
    # define GL here as a class attribute that includes all the opengl funcs
    # e.g. GL = pyglet.gl

    # define the name of the backend, used to register the name to use when
    # specifying `winType`
    # e.g. winTypeName = 'custom'

    def __init__(self, win):
        """Set up the backend window according the params of the PsychoPy win

        Before PsychoPy 1.90.0 this code was executed in Window._setupPyglet()

        :param: win is a PsychoPy Window (usually not fully created yet)
        """
        self.win = win  # this will use the @property to make/use a weakref
        pass

    def swapBuffers(self):
        """Set the gamma table for the graphics card

        """
        raise NotImplementedError(
                "Backend has failed to override a necessary method")

    @attributeSetter
    def gamma(self, gamma):
        """Set the gamma table for the graphics card

        :param gamma: a single value or a triplet for separate RGB gamma values
        """
        self.__dict__['gamma'] = gamma
        raise NotImplementedError(
                "Backend has failed to override a necessary method")

    @attributeSetter
    def gammaRamp(self, gammaRamp):
        """Gets the gamma ramp or sets it to a new value (an Nx3 or Nx1 array)
        """
        self.__dict__['gammaRamp'] = gammaRamp
        raise NotImplementedError(
                "Backend has failed to override a necessary method")

    @property
    def shadersSupported(self):
        """This is a read-only property indicating whether or not this backend
        supports OpenGL shaders"""
        raise NotImplementedError(
                "Backend has failed to override a necessary method")

    def setMouseVisibility(self, visibility):
        """Set visibility of the mouse to True or False"""
        raise NotImplementedError(
                "Backend has failed to override a necessary method")

    # Optional, depending on backend needs

    def dispatchEvents(self):
        """This method is not needed for all backends but for engines with an
        event loop it may be needed to pump for new events (e.g. pyglet)
        """
        logging.warning("dispatchEvents() method in {} was called "
                        "but is not implemented. Is it needed?"
                        .format(self.win.winType)
                        )

    def onResize(self, width, height):
        """A method that will be called if the window detects a resize event
        """
        logging.warning("dispatchEvents() method in {} was called "
                        "but is not implemented. Is it needed?"
                        .format(self.win.winType)
                        )

    def setCurrent(self):
        """Sets this window to be the current rendering target (for backends
        where 2 windows are permitted, e.g. not pygame)
        """
        pass

    # Helper methods that don't need converting

    @property
    def win(self):
        """The PsychoPy Window that this backend is supporting, which provides
        various important variables (like size, units, color etc).

        NB win is stored as a weakref to a psychopy.window and this property
        helpfully converts it back to a regular object so you don't need to
        think about it!
        """
        ref = self.__dict__['win']
        return ref()

    @win.setter
    def win(self, win):
        """The PsychoPy Window that this backend is supporting, which provides
        various important variables (like size, units, color etc).

        NB win is stored as a weakref to a psychopy.window and this property
        helpfully converts it back to a regular object so you don't need to
        think about it!
        """
        self.__dict__['win'] = weakref.ref(win)

    @property
    def autoLog(self):
        """If the window has logging turned on then backend should too"""
        return self.win.autoLog

    @property
    def name(self):
        """Name of the backend is only used for logging purposes"""
        return "{}_backend".format(self.win.name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Minolta light-measuring devices
See http://www.konicaminolta.com/instruments

----------
"""
from __future__ import absolute_import, print_function

from builtins import range
from builtins import object
from psychopy import logging
import struct
import sys
import time

try:
    import serial
except ImportError:
    serial = False


class LS100(object):
    """A class to define a Minolta LS100 (or LS110?) photometer

    You need to connect a LS100 to the serial (RS232) port and
    **when you turn it on press the F key** on the device. This will put
    it into the correct mode to communicate with the serial port.

    usage::

        from psychopy.hardware import minolta
        phot = minolta.LS100(port)
        if phot.OK:  # then we successfully made a connection
            print(phot.getLum())

    :parameters:

        port: string

            the serial port that should be checked

        maxAttempts: int
            If the device doesn't respond first time how many attempts
            should be made? If you're certain that this is the correct
            port and the device is on and correctly configured then this
            could be set high. If not then set this low.

    :troubleshooting:

        Various messages are printed to the log regarding the function
        of this device, but to see them you need to set the printing of
        the log to the correct level::

            from psychopy import logging
            logging.console.setLevel(logging.ERROR)  # error messages only
            logging.console.setLevel(logging.INFO)  # more info
            logging.console.setLevel(logging.DEBUG)  # log all communications

        If you're using a keyspan adapter (at least on macOS) be aware that
        it needs a driver installed. Otherwise no ports will be found.

        Error messages:

        ``ERROR: Couldn't connect to Minolta LS100/110 on ____``:
            This likely means that the device is not connected to that port
            (although the port has been found and opened). Check that the
            device has the `[` in the bottom right of the display;
            if not turn off and on again holding the `F` key.

        ``ERROR: No reply from LS100``:
            The port was found, the connection was made and an initial
            command worked, but then the device stopped communating. If the
            first measurement taken with the device after connecting does
            not yield a reasonable intensity the device can sulk (not a
            technical term!). The "[" on the display will disappear and you
            can no longer communicate with the device. Turn it off and on
            again (with F depressed) and use a reasonably bright screen for
            your first measurement. Subsequent measurements can be dark
            (or we really would be in trouble!!).
    """

    longName = "Minolta LS100/LS110"
    driverFor = ["ls110", "ls100"]

    def __init__(self, port, maxAttempts=1):
        super(LS100, self).__init__()

        if not serial:
            raise ImportError("The module serial is needed to connect to "
                              "photometers. On most systems this can be "
                              "installed with\n\t easy_install pyserial")

        if type(port) in [int, float]:
            # add one so that port 1=COM1
            self.portNumber = port
            self.portString = 'COM%i' % self.portNumber
        else:
            self.portString = port
            self.portNumber = None
        self.isOpen = 0
        self.lastQual = 0
        self.lastLum = None
        self.type = 'LS100'
        self.com = False
        self.OK = True  # until we fail
        self.maxAttempts = maxAttempts

        self.codes = {
            'ER00\r\n': 'Unknown command',
            'ER01\r\n': 'Setting error',
            'ER11\r\n': 'Memory value error',
            'ER10\r\n': 'Measuring range over',
            'ER19\r\n': 'Display range over',
            'ER20\r\n': 'EEPROM error (the photometer needs repair)',
            'ER30\r\n': 'Photometer battery exhausted', }

        # try to open the port
        _linux = sys.platform.startswith('linux')
        if sys.platform in ('darwin', 'win32') or _linux:
            try:
                self.com = serial.Serial(self.portString)
            except Exception:
                msg = ("Couldn't connect to port %s. Is it being used by "
                       "another program?")
                self._error(msg % self.portString)
        else:
            msg = "I don't know how to handle serial ports on %s"
            self._error(msg % sys.platform)
        # setup the params for comms
        if self.OK:
            self.com.close()  # not sure why this helps but on win32 it does!!
            # this is a slightly odd characteristic of the Minolta LS100
            self.com.bytesize = 7
            self.com.baudrate = 4800
            self.com.parity = serial.PARITY_EVEN  # none
            self.com.stopbits = serial.STOPBITS_TWO
            try:
                if not self.com.isOpen():
                    self.com.open()
            except Exception:
                msg = "Opened serial port %s, but couldn't connect to LS100"
                self._error(msg % self.portString)
            else:
                self.isOpen = 1

        if self.OK:  # we have an open com port. try to send a command
            for repN in range(self.maxAttempts):
                time.sleep(0.2)
                for n in range(10):
                    # set to use absolute measurements
                    reply = self.sendMessage(b'MDS,04')
                    if reply[0:2] == 'OK':
                        self.OK = True
                        break
                    elif reply not in self.codes:
                        self.OK = False
                        break  # wasn't valid
                    else:
                        self.OK = False  # false so far but keep trying
        if self.OK:  # we have successfully sent and read a command
            logging.info("Successfully opened %s" % self.portString)

    def setMode(self, mode='04'):
        """Set the mode for measurements. Returns True (success) or False

        '04' means absolute measurements.
        '08' = peak
        '09' = cont

        See user manual for other modes
        """
        reply = self.sendMessage(b'MDS,%s' % mode)
        return self.checkOK(reply)

    def measure(self):
        """Measure the current luminance and set .lastLum to this value
        """
        reply = self.sendMessage(b'MES')
        if self.checkOK(reply):
            lum = float(reply.split()[-1])
            return lum
        else:
            return -1

    def getLum(self):
        """Makes a measurement and returns the luminance value
        """
        return self.measure()

    def clearMemory(self):
        """Clear the memory of the device from previous measurements
        """
        reply = self.sendMessage(b'CLE')
        ok = self.checkOK(reply)
        return ok

    def checkOK(self, msg):
        """Check that the message from the photometer is OK.
        If there's an error show it (printed).

        Then return True (OK) or False.
        """
        # also check that the reply is what was expected
        if msg[0:2] != 'OK':
            if msg == '':
                logging.error('No reply from LS100')
                sys.stdout.flush()
            else:
                logging.error('Error message from LS100:' + self.codes[msg])
                sys.stdout.flush()
            return False
        else:
            return True

    def sendMessage(self, message, timeout=5.0):
        """Send a command to the photometer and wait an allotted
        timeout for a response.
        """
        if message[-2:] != '\r\n':
            message += '\r\n'  # append a newline if necess

        # flush the read buffer first
        # read as many chars as are in the buffer
        self.com.read(self.com.inWaiting()).encode('utf-8')
        for attemptN in range(self.maxAttempts):
            # send the message
            time.sleep(0.1)
            if type(message)!=bytes:
                message = message.decode('utf-8')
            self.com.write(message)
            self.com.flush()
            time.sleep(0.1)
            # get reply (within timeout limit)
            self.com.timeout = timeout
            # send complete message
            logging.debug('Sent command:' + message[:-2])
            retVal = self.com.readline().encode('utf-8')
            if len(retVal) > 0:
                break  # we got a reply so can stop trying

        return retVal

    def _error(self, msg):
        self.OK = False
        logging.error(msg)

    def setMaxAttempts(self, maxAttempts):
        """Changes the number of attempts to send a message and read the
        output. Typically this should be low initially, if you aren't sure
        that the device is setup correctly but then, after the first
        successful reading, set it higher.
        """
        self.maxAttempts = maxAttempts



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo and utility for finding out the key-code for specific keys
"""

from __future__ import absolute_import, division, print_function

from psychopy import visual, event, core

win = visual.Window([400, 400])
msg = visual.TextStim(win, text='press a key\n < esc > to quit')
msg.draw()
win.flip()

k = ['']
count = 0
while k[0] not in ['escape', 'esc'] and count < 5:
    k = event.waitKeys()
    print(k)
    count += 1

win.close()
core.quit()

# The contents of this file are in the public domain.



# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
from __future__ import division, absolute_import

import json
import os
import sys
import tempfile
import time

import gevent
import psutil

try:
    import iohub
    from iohub import IOHUB_DIRECTORY, _ispkg
    from iohub.devices import Computer
    Computer.is_iohub_process = True
    from iohub.errors import printExceptionDetailsToStdErr
    from iohub.server import ioServer
    from iohub import util
    from iohub.util import updateDict, yload, yLoader
except ImportError:
    import psychopy.iohub
    from psychopy.iohub import IOHUB_DIRECTORY, _ispkg
    from psychopy.iohub.devices import Computer

    Computer.is_iohub_process = True
    from psychopy.iohub.errors import printExceptionDetailsToStdErr
    from psychopy.iohub.server import ioServer
    from psychopy.iohub import util
    from psychopy.iohub.util import updateDict, yload, yLoader


def run(rootScriptPathDir, configFilePath):
    s = None
    try:
        if _ispkg:
            iohub.EXP_SCRIPT_DIRECTORY = rootScriptPathDir
        else:
            psychopy.iohub.EXP_SCRIPT_DIRECTORY = rootScriptPathDir

        tdir = tempfile.gettempdir()
        cdir, _ = os.path.split(configFilePath)
        if tdir == cdir:
            tf = open(configFilePath)
            ioHubConfig = json.loads(tf.read())
            tf.close()
            os.remove(configFilePath)
        else:
            ioHubConfig = yload(open(configFilePath, 'r'), Loader=yLoader)

        hub_config_path = os.path.join(IOHUB_DIRECTORY, 'default_config.yaml')

        hub_defaults_config = yload(open(hub_config_path, 'r'), Loader=yLoader)
        updateDict(ioHubConfig, hub_defaults_config)

        s = ioServer(rootScriptPathDir, ioHubConfig)
        udp_port = s.config.get('udp_port', 9000)
        s.log("Receiving diagram's on: {}".format(udp_port))
        s.udpService.start()
        s.setStatus("INITIALIZING")
        msgpump_interval = s.config.get('windows_msgpump_interval', 0.005)
        glets = []

        tlet = gevent.spawn(s.pumpMsgTasklet, msgpump_interval)
        glets.append(tlet)
        for m in s.deviceMonitors:
            m.start()
            glets.append(m)

        tlet = gevent.spawn(s.processEventsTasklet, 0.01)
        glets.append(tlet)

        if Computer.psychopy_process:
            tlet = gevent.spawn(s.checkForPsychopyProcess, 0.5)
            glets.append(tlet)

        s.setStatus("RUNNING")

        if hasattr(gevent, 'run'):
            gevent.run()
            glets = []
        else:
            gevent.joinall(glets)

        lrtime = Computer.global_clock.getLastResetTime()
        s.log('Server END Time Offset: {0}'.format(lrtime), 'DEBUG')
        return True

    except Exception: # pylint: disable=broad-except
        printExceptionDetailsToStdErr()
        if s:
            s.shutdown()
        return False

if __name__ == '__main__':
    psychopy_pid = None
    initial_offset = 0.0
    scriptPathDir = None
    configFileName = None

    prog = sys.argv[0]

    if len(sys.argv) >= 2:
        initial_offset = float(sys.argv[1])
    if len(sys.argv) >= 3:
        scriptPathDir = sys.argv[2]
    if len(sys.argv) >= 4:
        configFileName = sys.argv[3]
    if len(sys.argv) >= 5:
        psychopy_pid = int(sys.argv[4])
    if len(sys.argv) < 2:
        psychopy_pid = None
        configFileName = None
        scriptPathDir = None
        initial_offset = Computer.getTime()

    if psychopy_pid:
        Computer.psychopy_process = psutil.Process(psychopy_pid)
    Computer.global_clock = util.clock.MonotonicClock(initial_offset)

    run(rootScriptPathDir=scriptPathDir, configFilePath=configFileName)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Functions and classes related to file and directory error handling
"""
from __future__ import absolute_import, print_function

from builtins import str
import os
import glob

from psychopy import logging


def handleFileCollision(fileName, fileCollisionMethod):
    """Handle filename collisions by overwriting, renaming, or failing hard.

    :Parameters:

        fileCollisionMethod: 'overwrite', 'rename', 'fail'
            If a file with the requested name already exists, specify
            how to deal with it. 'overwrite' will overwite existing
            files in place, 'rename' will append an integer to create
            a new file ('trials1.psydat', 'trials2.pysdat' etc) and
            'error' will raise an IOError.
    """
    if fileCollisionMethod == 'overwrite':
        logging.warning('Data file, %s, will be overwritten' % fileName)
    elif fileCollisionMethod == 'fail':
        msg = ("Data file %s already exists. Set argument "
               "fileCollisionMethod to overwrite.")
        raise IOError(msg % fileName)
    elif fileCollisionMethod == 'rename':
        rootName, extension = os.path.splitext(fileName)
        matchingFiles = glob.glob("%s*%s" % (rootName, extension))

        # Build the renamed string.
        if not matchingFiles:
            fileName = "%s%s" % (rootName, extension)
        else:
            fileName = "%s_%d%s" % (rootName, len(matchingFiles), extension)

        # Check to make sure the new fileName hasn't been taken too.
        if os.path.exists(fileName):
            msg = ("New fileName %s has already been taken. Something "
                   "is wrong with the append counter.")
            raise IOError(msg % fileName)

    else:
        msg = "Argument fileCollisionMethod was invalid: %s"
        raise ValueError(msg % str(fileCollisionMethod))

    return fileName



"""
ioHub Common Eye Tracker Interface for Tobii (C) Eye Tracking System.
Calibration graphics implemented using PsychoPy.
"""
# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).

import psychopy
from psychopy import visual
import gevent
import time
import copy
import numpy as np
from collections import OrderedDict
from .....util import convertCamelToSnake
from .... import DeviceEvent, Computer
from .....constants import EventConstants
from .....errors import print2err, printExceptionDetailsToStdErr
currentTime = Computer.getTime


class TobiiPsychopyCalibrationGraphics(object):
    IOHUB_HEARTBEAT_INTERVAL = 0.050   # seconds between forced run through of
    # micro threads, since one is blocking
    # on camera setup.
    WINDOW_BACKGROUND_COLOR = (128, 128, 128)
    CALIBRATION_POINT_LIST = [
        (0.5, 0.5), (0.1, 0.1), (0.9, 0.1), (0.9, 0.9), (0.1, 0.9), (0.5, 0.5)]

    TEXT_POS = [0, 0]
    TEXT_COLOR = [0, 0, 0]
    TEXT_HEIGHT = 36
    _keyboard_key_index = EventConstants.getClass(
        EventConstants.KEYBOARD_RELEASE).CLASS_ATTRIBUTE_NAMES.index('key')

    def __init__(self, eyetrackerInterface, screenColor=None,
                 calibrationPointList=None):
        self._eyetrackerinterface = eyetrackerInterface

        self._tobii = eyetrackerInterface._tobii
        self.screenSize = eyetrackerInterface._display_device.getPixelResolution()

        self.width = self.screenSize[0]
        self.height = self.screenSize[1]
        self._ioKeyboard = None

        self._msg_queue = []
        self._lastCalibrationOK = False

        TobiiPsychopyCalibrationGraphics.WINDOW_BACKGROUND_COLOR = screenColor

        if calibrationPointList is not None:
            TobiiPsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = calibrationPointList

        calibration_methods = dict(THREE_POINTS=3,
                                   FIVE_POINTS=5,
                                   NINE_POINTS=9,
                                   THIRTEEN_POINTS=13)

        cal_type = self._eyetrackerinterface.getConfiguration()['calibration'][
            'type']

        if cal_type in calibration_methods:
            num_points = calibration_methods[cal_type]

            if num_points == 3:
                TobiiPsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = [
                    (0.5, 0.1), (0.1, 0.9), (0.9, 0.9), (0.5, 0.1)]
            elif num_points == 9:
                TobiiPsychopyCalibrationGraphics.CALIBRATION_POINT_LIST = [(0.5, 0.5),
                                                                           (0.1,
                                                                            0.5),
                                                                           (0.9,
                                                                            0.5),
                                                                           (0.1,
                                                                            0.1),
                                                                           (0.5,
                                                                            0.1),
                                                                           (0.9,
                                                                            0.1),
                                                                           (0.9,
                                                                            0.9),
                                                                           (0.5,
                                                                            0.9),
                                                                           (0.1,
                                                                            0.9),
                                                                           (0.5, 0.5)]
        display = self._eyetrackerinterface._display_device
        self.window = visual.Window(
            self.screenSize,
            monitor=display.getPsychopyMonitorName(),
            units=display.getCoordinateType(),
            fullscr=True,
            allowGUI=False,
            screen=display.getIndex(),
            color=self.WINDOW_BACKGROUND_COLOR[
                0:3],
            colorSpace='rgb255')
        self.window.flip(clearBuffer=True)

        self._createStim()
        self._registerEventMonitors()
        self._lastMsgPumpTime = currentTime()

        self.clearAllEventBuffers()

    def clearAllEventBuffers(self):
        self._eyetrackerinterface._iohub_server.eventBuffer.clear()
        for d in self._eyetrackerinterface._iohub_server.devices:
            d.clearEvents()

    def _registerEventMonitors(self):
        if self._eyetrackerinterface._iohub_server:
            for dev in self._eyetrackerinterface._iohub_server.devices:
                #ioHub.print2err("dev: ",dev.__class__.__name__)
                if dev.__class__.__name__ == 'Keyboard':
                    kbDevice = dev

        if kbDevice:
            eventIDs = []
            for event_class_name in kbDevice.__class__.EVENT_CLASS_NAMES:
                eventIDs.append(
                    getattr(
                        EventConstants,
                        convertCamelToSnake(
                            event_class_name[
                                :-5],
                            False)))

            self._ioKeyboard = kbDevice
            self._ioKeyboard._addEventListener(self, eventIDs)
        else:
            print2err(
                'Warning: Tobii Cal GFX could not connect to Keyboard device for events.')

    def _unregisterEventMonitors(self):
        if self._ioKeyboard:
            self._ioKeyboard._removeEventListener(self)

    def _handleEvent(self, event):
        event_type_index = DeviceEvent.EVENT_TYPE_ID_INDEX
        if event[event_type_index] == EventConstants.KEYBOARD_PRESS:
            ek = event[self._keyboard_key_index]
            if isinstance(ek, bytes):
                ek = ek.decode('utf-8')
            if ek == ' ':
                self._msg_queue.append('SPACE_KEY_ACTION')
                self.clearAllEventBuffers()
            elif ek == 'escape':
                self._msg_queue.append('QUIT')
                self.clearAllEventBuffers()

    def MsgPump(self):
        # keep the psychopy window happy ;)
        if currentTime() - self._lastMsgPumpTime > self.IOHUB_HEARTBEAT_INTERVAL:
            # try to keep ioHub from being blocked. ;(
            if self._eyetrackerinterface._iohub_server:
                for dm in self._eyetrackerinterface._iohub_server.deviceMonitors:
                    dm.device._poll()
                self._eyetrackerinterface._iohub_server.processDeviceEvents()
            self._lastMsgPumpTime = currentTime()

    def getNextMsg(self):
        if len(self._msg_queue)>0:
            msg = self._msg_queue[0]
            self._msg_queue = self._msg_queue[1:]
            return msg

    def _createStim(self):
        """
            outer_diameter: 35
            outer_stroke_width: 5
            outer_fill_color: [255,255,255]
            outer_line_color: [255,255,255]
            inner_diameter: 5
            inner_stroke_width: 0
            inner_color: [0,0,0]
            inner_fill_color: [0,0,0]
            inner_line_color: [0,0,0]
            calibration_prefs=self._eyetrackerinterface.getConfiguration()['calibration']['target_attributes']
        """
        coord_type = self._eyetrackerinterface._display_device.getCoordinateType()
        calibration_prefs = self._eyetrackerinterface.getConfiguration()['calibration'][
            'target_attributes']
        self.calibrationPointOUTER = visual.Circle(
            self.window,
            pos=(
                0,
                0),
            lineWidth=calibration_prefs['outer_stroke_width'],
            radius=calibration_prefs['outer_diameter'] / 2.0,
            name='CP_OUTER',
            fillColor=calibration_prefs['outer_fill_color'],
            lineColor=calibration_prefs['outer_line_color'],
            fillColorSpace='rgb255',
            lineColorSpace='rgb255',
            opacity=1.0,
            interpolate=False,
            edges=64,
            units=coord_type)

        self.calibrationPointINNER = visual.Circle(
            self.window,
            pos=(
                0,
                0),
            lineWidth=calibration_prefs['inner_stroke_width'],
            radius=calibration_prefs['inner_diameter'] / 2.0,
            name='CP_INNER',
            fillColor=calibration_prefs['inner_fill_color'],
            lineColor=calibration_prefs['inner_line_color'],
            fillColorSpace='rgb255',
            lineColorSpace='rgb255',
            opacity=1.0,
            interpolate=False,
            edges=64,
            units=coord_type)

        instuction_text = 'Press SPACE to Start Calibration; ESCAPE to Exit.'
        self.textLineStim = visual.TextStim(self.window,
                                            text=instuction_text,
                                            pos=self.TEXT_POS,
                                            height=self.TEXT_HEIGHT,
                                            color=self.TEXT_COLOR,
                                            colorSpace='rgb255',
                                            units='pix',
                                            wrapWidth=self.width * 0.9)

        # create Tobii eye position feedback graphics
        #
        sw, sh = self.screenSize
        self.hbox_bar_length = hbox_bar_length = sw / 4
        hbox_bar_height = 6
        marker_diameter = 7
        self.marker_heights = (-sh / 2.0 * .7, -sh / 2.0 * .75, -sh /
                               2.0 * .8, -sh / 2.0 * .7, -sh / 2.0 * .75, -sh / 2.0 * .8)

        bar_vertices = [-hbox_bar_length / 2, -hbox_bar_height / 2], [hbox_bar_length / 2, -hbox_bar_height /
                                                                      2], [hbox_bar_length / 2, hbox_bar_height / 2], [-hbox_bar_length / 2, hbox_bar_height / 2]

        self.feedback_resources = OrderedDict()

        self.feedback_resources['hbox_bar_x'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='Firebrick',
            vertices=bar_vertices,
            pos=(
                0,
                self.marker_heights[0]))
        self.feedback_resources['hbox_bar_y'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='DarkSlateGray',
            vertices=bar_vertices,
            pos=(
                0,
                self.marker_heights[1]))
        self.feedback_resources['hbox_bar_z'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='GoldenRod',
            vertices=bar_vertices,
            pos=(
                0,
                self.marker_heights[2]))

        marker_vertices = [-marker_diameter, 0], [0,
                                                  marker_diameter], [marker_diameter, 0], [0, -marker_diameter]
        self.feedback_resources['left_hbox_marker_x'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='Black',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[0]))
        self.feedback_resources['left_hbox_marker_y'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='Black',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[1]))
        self.feedback_resources['left_hbox_marker_z'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='Black',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[2]))
        self.feedback_resources['right_hbox_marker_x'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='DimGray',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[0]))
        self.feedback_resources['right_hbox_marker_y'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='DimGray',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[1]))
        self.feedback_resources['right_hbox_marker_z'] = visual.ShapeStim(
            win=self.window,
            lineColor='White',
            fillColor='DimGray',
            vertices=marker_vertices,
            pos=(
                0,
                self.marker_heights[2]))

    def runCalibration(self):
        """Performs a simple Tobii - like (@2010) calibration routine.

        Args:
            None

        Result:
            bool: True if calibration was successful. False if not.

        """

        self._lastCalibrationOK = False
        calibration_sequence_completed = False

        instuction_text = 'Press SPACE to Start Calibration; ESCAPE to Exit.'
        continue_calibration = self.showSystemSetupMessageScreen(
            instuction_text, True)
        if not continue_calibration:
            return False

        auto_pace = self._eyetrackerinterface.getConfiguration()['calibration'][
            'auto_pace']
        pacing_speed = self._eyetrackerinterface.getConfiguration()['calibration'][
            'pacing_speed']
        randomize_points = self._eyetrackerinterface.getConfiguration()['calibration'][
            'randomize']

        cal_target_list = self.CALIBRATION_POINT_LIST[1:-1]
        if randomize_points is True:
            import random
            random.seed(None)
            random.shuffle(cal_target_list)

        cal_target_list.insert(0, self.CALIBRATION_POINT_LIST[0])
        cal_target_list.append(self.CALIBRATION_POINT_LIST[-1])

        calibration = self._tobii.newScreenCalibration()

        calibration.enter_calibration_mode()

        i = 0
        for pt in cal_target_list:
            self.clearAllEventBuffers()
            left, top, right, bottom = self._eyetrackerinterface._display_device.getCoordBounds()
            w, h = right - left, top - bottom
            x, y = left + w * pt[0], bottom + h * (1.0 - pt[1])
            self.drawCalibrationTarget(i, (x, y))
            self.clearAllEventBuffers()
            stime = currentTime()
            def waitingForNextTargetTime():
                return True

            if auto_pace is True:
                def waitingForNextTargetTime():
                    return currentTime() - stime < float(pacing_speed)

            _quit = False
            while waitingForNextTargetTime():
                msg = self.getNextMsg()
                if msg == 'SPACE_KEY_ACTION':
                    break
                elif msg == 'QUIT':
                    _quit = True
                    break
                self.MsgPump()
                gevent.sleep(0.01)

            calibration.collect_data(pt[0], pt[1])

            if _quit:
                calibration.leave_calibration_mode()
                calibration = None
                break

            # TODO: Switch to support per target status check available
            # in tobii_research.
            #if calibration.collect_data(pt[0], pt[1]) != self._tobii.CALIBRATION_STATUS_SUCCESS:
                # Try again if it didn't go well the first time.
                # Not all eye tracker models will fail at this point, but instead fail on ComputeAndApply.
            #    print2err("Calibration failed for target {}. Recollecting data.... TODO: Give visual feedback.")
            #    calibration.collect_data(pt[0], pt[1])

            self.clearCalibrationWindow()
            self.clearAllEventBuffers()

            i += 1
            if i == len(cal_target_list):
                calibration_sequence_completed = True

        self.clearCalibrationWindow()
        self.clearAllEventBuffers()

        if _quit:
            return False
        
        self._lastCalibrationOK = False
        if calibration:
            if calibration_sequence_completed:
                calibration_result = calibration.compute_and_apply()
                self._lastCalibrationOK = calibration_result.status == 'calibration_status_success'
            else:
                self._lastCalibrationOK = False
            calibration.leave_calibration_mode()
            calibration = None
            
            
            
        if self._lastCalibrationOK is False:
            instuction_text = 'Calibration Failed. Options: SPACE: Re-run Calibration; ESCAPE: Exit Setup'
            continue_method = self.showSystemSetupMessageScreen(
                instuction_text, True, msg_types=['SPACE_KEY_ACTION', 'QUIT'])
            if continue_method is False:
                return self.runCalibration()
            return False
        
        instuction_text = "Calibration Passed. PRESS 'SPACE' KEY TO CONTINUE."
        self.showSystemSetupMessageScreen(instuction_text, True, msg_types=['SPACE_KEY_ACTION'])
        return True
    
    def clearCalibrationWindow(self):
        self.window.flip(clearBuffer=True)

    def showSystemSetupMessageScreen(
        self,
        text_msg='Press SPACE to Start Calibration; ESCAPE to Exit.',
        enable_recording=False,
        msg_types=[
            'SPACE_KEY_ACTION',
            'QUIT']):
        if enable_recording is True:
            self._eyetrackerinterface.setRecordingState(True)

        self.clearAllEventBuffers()

        while True:
            self.textLineStim.setText(text_msg)
            event_named_tuples = []
            for e in self._eyetrackerinterface.getEvents(
                    EventConstants.BINOCULAR_EYE_SAMPLE):
                event_named_tuples.append(EventConstants.getClass(
                    EventConstants.BINOCULAR_EYE_SAMPLE).createEventAsNamedTuple(e))
            leye_box_pos, reye_box_pos = self.getHeadBoxPosition(
                event_named_tuples)
            lx, ly, lz = leye_box_pos
            rx, ry, rz = reye_box_pos
            eye_positions = (lx, ly, lz, rx, ry, rz)
            marker_names = (
                'left_hbox_marker_x',
                'left_hbox_marker_y',
                'left_hbox_marker_z',
                'right_hbox_marker_x',
                'right_hbox_marker_y',
                'right_hbox_marker_z')
            marker_heights = self.marker_heights
            hbox_bar_length = self.hbox_bar_length

            for i, p in enumerate(eye_positions):
                if p is not None:
                    mpoint = hbox_bar_length * p - \
                        hbox_bar_length / 2.0, marker_heights[i]
                    self.feedback_resources[marker_names[i]].setPos(mpoint)
                    self.feedback_resources[marker_names[i]].setOpacity(1.0)
                else:
                    self.feedback_resources[marker_names[i]].setOpacity(0.0)

            self.textLineStim.draw()
            [r.draw() for r in self.feedback_resources.values()]
            self.window.flip()

            msg = self.getNextMsg()
            if msg == 'SPACE_KEY_ACTION':
                if enable_recording is True:
                    self._eyetrackerinterface.setRecordingState(False)
                self.clearAllEventBuffers()
                return True
            elif msg == 'QUIT':
                if enable_recording is True:
                    self._eyetrackerinterface.setRecordingState(False)
                self.clearAllEventBuffers()
                return False
            self.MsgPump()
            gevent.sleep()

    def getHeadBoxPosition(self, events):
        # KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index('key_id')
        left_eye_cam_x = None
        left_eye_cam_y = None
        left_eye_cam_z = None
        right_eye_cam_x = None
        right_eye_cam_y = None
        right_eye_cam_z = None

        if len(events) == 0:
            return (left_eye_cam_x, left_eye_cam_y,
                    left_eye_cam_z), (right_eye_cam_x, right_eye_cam_y, right_eye_cam_z)

        event = events[-1]
        if abs(
                event.left_eye_cam_x) != 1.0 and abs(
                event.left_eye_cam_y) != 1.0:
            left_eye_cam_x = 1.0 - event.left_eye_cam_x
            left_eye_cam_y = event.left_eye_cam_y
        if event.left_eye_cam_z != 0.0:
            left_eye_cam_z = event.left_eye_cam_z
        if abs(
                event.right_eye_cam_x) != 1.0 and abs(
                event.right_eye_cam_y) != 1.0:
            right_eye_cam_x = 1.0 - event.right_eye_cam_x
            right_eye_cam_y = event.right_eye_cam_y
        if event.right_eye_cam_z != 0.0:
            right_eye_cam_z = event.right_eye_cam_z
        return (left_eye_cam_x, left_eye_cam_y,
                left_eye_cam_z), (right_eye_cam_x, right_eye_cam_y, right_eye_cam_z)

    def setTargetDefaults(self):
        """
            outer_diameter: 35
            outer_stroke_width: 5
            outer_fill_color: [255,255,255]
            outer_line_color: [255,255,255]
            inner_diameter: 5
            inner_stroke_width: 0
            inner_color: [0,0,0]
            inner_fill_color: [0,0,0]
            inner_line_color: [0,0,0]
            calibration_prefs=self._eyetrackerinterface.getConfiguration()['calibration']['target_attributes']
        """
        calibration_prefs = self._eyetrackerinterface.getConfiguration()['calibration'][
            'target_attributes']

        self.calibrationPointOUTER.radius = calibration_prefs[
            'outer_diameter'] / 2.0
        self.calibrationPointOUTER.setLineColor(
            calibration_prefs['outer_line_color'])
        self.calibrationPointOUTER.setFillColor(
            calibration_prefs['outer_fill_color'])
        self.calibrationPointOUTER.lineWidth = int(
            calibration_prefs['outer_stroke_width'])

        self.calibrationPointINNER.radius = calibration_prefs[
            'inner_diameter'] / 2.0
        self.calibrationPointINNER.setLineColor(
            calibration_prefs['inner_line_color'])
        self.calibrationPointINNER.setFillColor(
            calibration_prefs['inner_fill_color'])
        self.calibrationPointINNER.lineWidth = int(
            calibration_prefs['inner_stroke_width'])

        self.calibrationPointOUTER.draw()
        self.calibrationPointINNER.draw()
        return self.window.flip(clearBuffer=True)

    def moveTarget(self, start_pt, end_pt, TARG_VELOCITY):
        sx, sy = start_pt
        ex, ey = end_pt
        dist = np.linalg.norm(end_pt - start_pt)
        sec_dur = dist / TARG_VELOCITY
        num_retraces = sec_dur / self._eyetrackerinterface._display_device.getRetraceInterval()
        x_points = np.linspace(sx, ex, num=int(num_retraces))
        y_points = np.linspace(sy, ey, num=int(num_retraces))
        t_points = zip(x_points, y_points)
        for p in t_points:
            self.calibrationPointOUTER.setPos(p)
            self.calibrationPointINNER.setPos(p)
            self.calibrationPointOUTER.draw()
            self.calibrationPointINNER.draw()
            self.window.flip(clearBuffer=True)
        self.setTargetDefaults()

    def expandTarget(self, TARG_RAD_MULTIPLIER, EXPANSION_RATE):
        calibration_prefs = self._eyetrackerinterface.getConfiguration()['calibration'][
            'target_attributes']
        orad = calibration_prefs['outer_diameter'] / 2.0
        self.calibrationPointOUTER.lineWidth = int(
            calibration_prefs['outer_stroke_width'])
        if self.calibrationPointOUTER.lineWidth < 1:
            self.calibrationPointOUTER.lineWidth = 1

        max_osize = orad * TARG_RAD_MULTIPLIER
        if EXPANSION_RATE < 1:
            EXPANSION_RATE = 1.0

        stime = Computer.getTime()
        self.calibrationPointOUTER.radius = orad
        self.calibrationPointOUTER.draw()
        self.calibrationPointINNER.draw()
        ftime = self.window.flip(clearBuffer=True)
        current_size = self.calibrationPointOUTER.radius
        while current_size < max_osize:
            sec_dur = ftime - stime
            if sec_dur < 0.0:
                sec_dur = 0.0
            stime = ftime
            current_size += sec_dur * EXPANSION_RATE
            self.calibrationPointOUTER.radius = current_size
            self.calibrationPointOUTER.draw()
            self.calibrationPointINNER.draw()
            ftime = self.window.flip(clearBuffer=True)

    def contractTarget(self, TARG_RAD_MULTIPLIER, EXPANSION_RATE):
        calibration_prefs = self._eyetrackerinterface.getConfiguration()['calibration'][
            'target_attributes']
        orad = calibration_prefs['outer_diameter'] / 2.0
        self.calibrationPointOUTER.lineWidth = int(
            calibration_prefs['outer_stroke_width'])
        if self.calibrationPointOUTER.lineWidth < 1:
            self.calibrationPointOUTER.lineWidth = 1

        max_osize = orad * TARG_RAD_MULTIPLIER
        if EXPANSION_RATE < 1:
            EXPANSION_RATE = 1.0

        stime = Computer.getTime()
        self.calibrationPointOUTER.radius = max_osize
        self.calibrationPointOUTER.draw()
        self.calibrationPointINNER.draw()
        ftime = self.window.flip(clearBuffer=True)
        current_size = max_osize
        while current_size > orad:
            sec_dur = ftime - stime
            if sec_dur < 0.0:
                sec_dur = 0.0
            stime = ftime
            current_size -= sec_dur * EXPANSION_RATE
            self.calibrationPointOUTER.radius = current_size
            self.calibrationPointOUTER.draw()
            self.calibrationPointINNER.draw()
            ftime = self.window.flip(clearBuffer=True)

    def drawCalibrationTarget(self, target_number, tp):
        """
            outer_diameter: 35
            outer_stroke_width: 5
            outer_fill_color: [255,255,255]
            outer_line_color: [255,255,255]
            inner_diameter: 5
            inner_stroke_width: 0
            inner_color: [0,0,0]
            inner_fill_color: [0,0,0]
            inner_line_color: [0,0,0]
            calibration_prefs=self._eyetrackerinterface.getConfiguration()['calibration']['target_attributes']
        """
        try:
            calibration_prefs = self._eyetrackerinterface.getConfiguration()['calibration'][
                'target_attributes']
            animate_prefs = calibration_prefs.get('animate', None)

            if animate_prefs:
                CONTRACT_ONLY = animate_prefs.get('contract_only', False)
                TARG_VELOCITY = animate_prefs.get(
                    'movement_velocity', 300.0)  # 200 pix / sec
                TARG_RAD_MULTIPLIER = animate_prefs.get('expansion_ratio', 3.0)
                EXPANSION_RATE = animate_prefs.get('expansion_speed', 30.0)

                if target_number == 0:
                    # Do first point animation
                    self.calibrationPointOUTER.setPos(tp)
                    self.calibrationPointINNER.setPos(tp)
                    self.setTargetDefaults()
                    if CONTRACT_ONLY is False:
                        self.expandTarget(TARG_RAD_MULTIPLIER, EXPANSION_RATE)
                    self.contractTarget(TARG_RAD_MULTIPLIER, EXPANSION_RATE)
                else:
                    # Move from current point to new point
                    # then do point animation
                    spos = self.calibrationPointOUTER.pos
                    # self.calibrationPointOUTER.setPos(tp)
                    # self.calibrationPointINNER.setPos(tp)
                    if TARG_VELOCITY > 0.0:
                        self.moveTarget(spos, tp, TARG_VELOCITY)
                    else:
                        self.calibrationPointOUTER.setPos(tp)
                        self.calibrationPointINNER.setPos(tp)
                    self.setTargetDefaults()
                    if CONTRACT_ONLY is False:
                        self.expandTarget(TARG_RAD_MULTIPLIER, EXPANSION_RATE)
                    self.contractTarget(TARG_RAD_MULTIPLIER, EXPANSION_RATE)
            else:
                self.calibrationPointOUTER.setPos(tp)
                self.calibrationPointINNER.setPos(tp)
                self.setTargetDefaults()

        except Exception:
            printExceptionDetailsToStdErr()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# We deliberately delay importing the pyparallel module until we try
# to use it - this allows us to import the class on machines
# which don't have it and then worry about dealing with
# using the right one later

# This is necessary to stop the local parallel.py masking the module
# we actually want to find!

from __future__ import absolute_import, print_function

# We duck-type the parallel port objects


from builtins import object
class PParallelLinux(object):
    """This class provides read/write access to the parallel port for linux
    using pyparallel.

    Note that you must have the lp module removed and the ppdev module loaded
    to use this code::
        sudo rmmod lp
        sudo modprobe ppdev
    """

    def __init__(self, address='/dev/parport0'):
        """Set the device node of your parallel port

        Common port addresses::

            LPT1 = /dev/parport0
            LPT2 = /dev/parport1
            LPT3 = /dev/parport2
        """
        import parallel as pyp

        if not hasattr(pyp, 'Parallel'):
            # We failed to import pyparallel properly
            # We probably ended up with psychopy.parallel instead...
            raise Exception('Failed to import pyparallel - is it installed?')

        self.port = pyp.Parallel(address)
        self.status = None

    def __del__(self):
        if hasattr(self, 'port'):
            del self.port

    def setData(self, data):
        """Set the data to be presented on the parallel port (one ubyte).
        Alternatively you can set the value of each pin (data pins are pins
        2-9 inclusive) using :func:`~psychopy.parallel.setPin`

        Examples::

            p.setData(0)  # sets all pins low
            p.setData(255)  # sets all pins high
            p.setData(2)  # sets just pin 3 high (remember that pin2=bit0)
            p.setData(3)  # sets just pins 2 and 3 high

        You can also convert base 2 to int easily in python::

            parallel.setData(int("00000011", 2))  # pins 2 and 3 high
            parallel.setData(int("00000101", 2))  # pins 2 and 4 high
        """
        self.port.setData(data)

    def setPin(self, pinNumber, state):
        """Set a desired pin to be high(1) or low(0).

        Only pins 2-9 (incl) are normally used for data output::

            p.setPin(3, 1)  # sets pin 3 high
            p.setPin(3, 0)  # sets pin 3 low
        """
        # I can't see how to do this without reading and writing the data
        if state:
            self.port.setData(self.port.PPRDATA() | (2**(pinNumber - 2)))
        else:
            self.port.setData(self.port.PPRDATA() & (255 ^ 2**(pinNumber - 2)))

    def readData(self):
        """Return the value currently set on the data pins (2-9)
        """
        return self.port.PPRDATA()

    def readPin(self, pinNumber):
        """Determine whether a desired (input) pin is high(1) or low(0).

        Pins 2-13 and 15 are currently read here
        """
        if pinNumber == 10:
            return self.port.getInAcknowledge()
        elif pinNumber == 11:
            return self.port.getInBusy()
        elif pinNumber == 12:
            return self.port.getInPaperOut()
        elif pinNumber == 13:
            return self.port.getInSelected()
        elif pinNumber == 15:
            return self.port.getInError()
        elif 2 <= pinNumber <= 9:
            return (self.port.PPRDATA() >> (pinNumber - 2)) & 1
        else:
            msg = 'Pin %i cannot be read (by PParallelLinux.readPin())'
            print(msg % pinNumber)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tools for the use with the :py:class:`~psychopy.visual.rift.Rift` class.

This module exposes additional useful classes and functions from PsychXR without
needing to explicitly import the PsychXR library into your project. If PsychXR
is not available on your system, class objects will be `None`.

Copyright (C) 2019 - Matthew D. Cutone, The Centre for Vision Research, Toronto,
Ontario, Canada

"""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

__all__ = ['LibOVRPose',
           'LibOVRPoseState',
           'LibOVRBounds',
           'LibOVRHapticsBuffer',
           'isHmdConnected',
           'isOculusServiceRunning']

_HAS_PSYCHXR_ = True

try:
    import psychxr.libovr as libovr
except ImportError:
    _HAS_PSYCHXR_ = False


LibOVRPose = libovr.LibOVRPose if _HAS_PSYCHXR_ else None
LibOVRPoseState = libovr.LibOVRPoseState if _HAS_PSYCHXR_ else None
LibOVRBounds = libovr.LibOVRBounds if _HAS_PSYCHXR_ else None
LibOVRHapticsBuffer = libovr.LibOVRHapticsBuffer if _HAS_PSYCHXR_ else None


def isHmdConnected(timeout=0):
    """Check if an HMD is connected.

    Parameters
    ----------
    timeout : int
        Timeout in milliseconds.

    Returns
    -------
    bool
        `True` if an HMD is connected.

    """
    if _HAS_PSYCHXR_:
        return libovr.isHmdConnected(timeout)

    return False


def isOculusServiceRunning(timeout=0):
    """Check if the Oculus(tm) service is currently running.

    Parameters
    ----------
    timeout : int
        Timeout in milliseconds.

    Returns
    -------
    bool
        `True` if the service is loaded and running.

    """
    if _HAS_PSYCHXR_:
        return libovr.isOculusServiceRunning(timeout)

    return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of class psychopy.visual.BufferImageStim()

- take a snapshot of a multi-item screen image
- save and draw as a BufferImageStim
- report speed of BufferImageStim to speed of drawing each item separately
"""

from __future__ import division
from __future__ import print_function

from builtins import str
from builtins import range
from psychopy import visual, event, core

# need a window and clock:
win = visual.Window(fullscr=False, monitor='testMonitor')
clock = core.Clock()

# first define a list of various slow, static stim
imageList = ['face.jpg', 'beach.jpg']
imageStim = visual.SimpleImageStim(win, imageList[0])
imageStim2 = visual.SimpleImageStim(win, imageList[1], pos=(.300, .20))
wordStim = visual.TextStim(win,
            text='Press < escape > to quit.\n\nThere should be no change after 3 seconds.\n\n' +
            'This is a text stim that is kinda verbose and long, so if it ' +
            'were actually really long it would take a while to render completely.',
            pos=(0, -.2))
stimlist = [imageStim, imageStim2, wordStim]

# Get and save a "screen shot" of everything in stimlist:
rect = (-1, 1, 1, -1)
t0 = clock.getTime()
screenshot = visual.BufferImageStim(win, stim=stimlist, rect=rect)
# rect is the screen rectangle to grab, (-1, 1, 1, -1) is whole-screen
# as a list of the edges: Left Top Right Bottom, in norm units.

captureTime = clock.getTime() - t0

instr_buffer = visual.TextStim(win, text='BufferImageStim', pos=(0, .8))
drawTimeBuffer = []  # accumulate draw times of the screenshot
for frameCounter in range(200):
    t0 = clock.getTime()
    screenshot.draw()  # draw the BufferImageStim, fast
    drawTimeBuffer.append(clock.getTime() - t0)
    instr_buffer.draw()
    win.flip()
    if len(event.getKeys(['escape'])):
        core.quit()

# Just for the demo: Time things when drawn individually:
instr_multi = visual.TextStim(win, text='TextStim and ImageStim', pos=(0, .8))
drawTimeMulti  = []  # draw times of the pieces, as drawn separately
for frameCounter in range(200):
    t0 = clock.getTime()
    for s in stimlist:
        s.draw()  # draw all individual stim, slow
    drawTimeMulti.append(clock.getTime() - t0)
    instr_multi.draw()
    win.flip()
    if len(event.getKeys(['escape'])):
        core.quit()

# Report timing:
firstFrameTime = drawTimeBuffer.pop(0)
bufferAvg = 1000. * sum(drawTimeBuffer) / len(drawTimeBuffer)
multiAvg = 1000. * sum(drawTimeMulti) / len(drawTimeMulti)
msg = "\nBufferImageStim\nrect=%s norm units, becomes %s pix"
print(msg % (str(rect), str(screenshot.size)))
print("initial set-up / screen capture: %.0fms total" % (1000. * captureTime))
print("first frame:    %.2fms (typically slow)" % (1000. * firstFrameTime))
msg = "BufferImage:    %.2fms avg, %.2fms max draw-time (%d frames)"
print(msg % (bufferAvg, max(drawTimeBuffer) * 1000., len(drawTimeBuffer)))
msg = "Text & Image:   %.2fms avg, %.2fms max draw-time"
print(msg % (multiAvg, max(drawTimeMulti) * 1000.))

win.close()
core.quit()

# The contents of this file are in the public domain.



# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
import serial
import sys
import numpy as N
import struct
from ... import EXP_SCRIPT_DIRECTORY
from .. import Device, DeviceEvent, Computer
from ...errors import print2err, printExceptionDetailsToStdErr
from ...constants import DeviceConstants, EventConstants
getTime = Computer.getTime

PY3 = sys.version_info.major >= 3


class Serial(Device):
    """A general purpose serial input interface device.

    Configuration options are used to define how the serial input data
    should be parsed, and what conditions create a serial input event to
    be generated.

    """
    _bytesizes = {
        5: serial.FIVEBITS,
        6: serial.SIXBITS,
        7: serial.SEVENBITS,
        8: serial.EIGHTBITS,
    }

    _parities = {
        'NONE': serial.PARITY_NONE,
        'EVEN': serial.PARITY_EVEN,
        'ODD': serial.PARITY_ODD,
        'MARK': serial.PARITY_MARK,
        'SPACE': serial.PARITY_SPACE
    }

    _stopbits = {
        'ONE': serial.STOPBITS_ONE,
        'ONE_AND_HALF': serial.STOPBITS_ONE_POINT_FIVE,
        'TWO': serial.STOPBITS_TWO
    }

    DEVICE_TIMEBASE_TO_SEC = 1.0
    _newDataTypes = [('port', '|S32'), ('baud', '|S32'), ]
    EVENT_CLASS_NAMES = ['SerialInputEvent', 'SerialByteChangeEvent']
    DEVICE_TYPE_ID = DeviceConstants.SERIAL
    DEVICE_TYPE_STRING = 'SERIAL'
    _serial_slots = [
        'port', 'baud', 'bytesize', 'parity', 'stopbits', '_serial',
        '_timeout', '_rx_buffer', '_parser_config', '_parser_state',
        '_event_count', '_byte_diff_mode', '_custom_parser',
        '_custom_parser_kwargs'
    ]
    __slots__ = [e for e in _serial_slots]

    def __init__(self, *args, **kwargs):
        Device.__init__(self, *args, **kwargs['dconfig'])
        self._serial = None
        self.port = self.getConfiguration().get('port')
        if self.port.lower() == 'auto':
            pports = self.findPorts()
            if pports:
                self.port = pports[0]
                if len(pports) > 1:
                    print2err(
                        'Warning: Serial device port configuration set '
                        "to 'auto'.\nMultiple serial ports found:\n",
                        pports, '\n** Using port ', self.port
                    )
        self.baud = self.getConfiguration().get('baud')
        self.bytesize = self._bytesizes[
            self.getConfiguration().get('bytesize')]
        self.parity = self._parities[self.getConfiguration().get('parity')]
        self.stopbits = self._stopbits[self.getConfiguration().get('stopbits')]

        self._parser_config = self.getConfiguration().get('event_parser')
        self._byte_diff_mode = None
        self._custom_parser = None
        self._custom_parser_kwargs = {}
        custom_parser_func_str = self._parser_config.get('parser_function')
        if custom_parser_func_str:
            # print2err("CUSTOM SERIAL PARSER FUNC STR: ", custom_parser_func_str)
            # Function referenced by string must have the following signature:
            #
            # evt_list = someCustomParserName(read_time, rx_data, parser_state, **kwargs)
            #
            # where:
            #     read_time: The time when the serial device read() returned
            #             with the new rx_data.
            #     rx_data: The new serial data received. Any buffering of data
            #             across function calls must be done by the function
            #             logic itself. parser_state could be used to hold
            #             such a buffer if needed.
            #     parser_state: A dict which can be used by the function to
            #             store any values that need to be accessed
            #             across multiple calls to the function. The dict
            #             is initially empty.
            #     kwargs: The parser_kwargs preference dict read from
            #             the event_parser preferences; or an empty dict if
            #             parser_kwargs was not found.
            #
            # The function must return a list like object, used to provide ioHub
            # with any new serial events that have been found.
            # Each element of the list must be a dict like object, representing
            # a single serial device event found by the parsing function.
            # A dict can contain the following key, value pairs:
            #    data: The string containing the parsed event data. (REQUIRED)
            #    time: The timestamp for the event (Optional). If not provided,
            #          the return time of the latest serial.read() is used.
            # Each event returned by the function generates a SerialInputEvent
            # with the data field = the dict data value.
            import importlib
            import sys
            try:
                #print2err("EXP_SCRIPT_DIRECTORY: ",EXP_SCRIPT_DIRECTORY)
                if EXP_SCRIPT_DIRECTORY not in sys.path:
                    sys.path.append(EXP_SCRIPT_DIRECTORY)
                mod_name, func_name = custom_parser_func_str.rsplit('.', 1)
                mod = importlib.import_module(mod_name)
                self._custom_parser = getattr(mod, func_name)
            except Exception:
                print2err(
                    'ioHub Serial Device Error: could not load '
                    'custom_parser function: ', custom_parser_func_str)
                printExceptionDetailsToStdErr()

            if self._custom_parser:
                self._custom_parser_kwargs = self._parser_config.get(
                    'parser_kwargs', {})
        else:
            self._byte_diff_mode = self._parser_config.get('byte_diff')

        if self._byte_diff_mode:
            self._rx_buffer = None
        else:
            self._resetParserState()
            self._rx_buffer = ''

        self._event_count = 0
        self._timeout = None
        self._serial = None
        self.setConnectionState(True)

    @classmethod
    def findPorts(cls):
        """Finds serial ports that are available."""
        import os
        available = []
        if os.name == 'nt':  # Windows
            for i in range(1, 256):
                try:
                    sport = 'COM' + str(i)
                    s = serial.Serial(sport)
                    available.append(sport)
                    s.close()
                except serial.SerialException:
                    pass
        else:  # Mac / Linux
            from serial.tools import list_ports
            available = [port[0] for port in list_ports.comports()]

        if len(available) < 1:
            print2err('Error: unable to find any serial ports on the computer.')
            return []
        return available

    def _resetParserState(self):
        parser_config = self._parser_config
        if parser_config:
            if self._custom_parser:
                self._parser_state = dict()
                return

            self._parser_state = dict(parsed_event='')
            parser_state = self._parser_state

            fixed_length = parser_config.setdefault('fixed_length', None)
            if fixed_length:
                parser_state['bytes_needed'] = fixed_length
            else:
                parser_state['bytes_needed'] = 0

            prefix = parser_config.setdefault('prefix', None)
            if prefix == r'\n':
                parser_config['prefix'] = '\n'
            elif prefix == r'\t':
                parser_config['prefix'] = '\t'
            elif prefix == r'\r':
                parser_config['prefix'] = '\r'
            elif prefix == r'\r\n':
                parser_config['prefix'] = '\r\n'

            if prefix:
                parser_state['prefix_found'] = False
            else:
                parser_state['prefix_found'] = True

            delimiter = parser_config.setdefault('delimiter', None)
            if delimiter == r'\n':
                parser_config['delimiter'] = '\n'
            elif delimiter == r'\t':
                parser_config['delimiter'] = '\t'
            elif delimiter == r'\r':
                parser_config['delimiter'] = '\r'
            elif delimiter == r'\r\n':
                parser_config['delimiter'] = '\r\n'
            if delimiter:
                parser_state['delimiter_found'] = False
            else:
                parser_state['delimiter_found'] = True

    def setConnectionState(self, enable):
        if enable is True:
            if self._serial is None:
                self._connectSerial()

        elif enable is False:
            if self._serial:
                self._serial.close()

        return self.isConnected()

    def isConnected(self):
        return self._serial is not None

    def getDeviceTime(self):
        return getTime()

    def getSecTime(self):
        """Returns current device time in sec.msec format.

        Relies on a functioning getDeviceTime() method.

        """
        return self.getTime()

    def enableEventReporting(self, enabled=True):
        """
        Specifies if the device should be reporting events to the ioHub Process
        (enabled=True) or whether the device should stop reporting events to the
        ioHub Process (enabled=False).


        Args:
            enabled (bool):  True (default) == Start to report device events to the ioHub Process. False == Stop Reporting Events to the ioHub Process. Most Device types automatically start sending events to the ioHUb Process, however some devices like the EyeTracker and AnlogInput device's do not. The setting to control this behavior is 'auto_report_events'

        Returns:
            bool: The current reporting state.
        """
        if enabled and not self.isReportingEvents():
            if not self.isConnected():
                self.setConnectionState(True)
            self.flushInput()
        if self._byte_diff_mode:
            self._rx_buffer = None
        else:
            self._rx_buffer = ''
        self._event_count = 0
        return Device.enableEventReporting(self, enabled)

    def isReportingEvents(self):
        """Returns whether a Device is currently reporting events to the ioHub
        Process.

        Args: None

        Returns:
            (bool): Current reporting state.

        """
        return Device.isReportingEvents(self)

    def _connectSerial(self):
        self._serial = serial.Serial(
            self.port, self.baud, timeout=self._timeout)
        if self._serial is None:
            raise ValueError(
                'Error: Serial Port Connection Failed: %d' %
                (self.port))
        self._serial.flushInput()
        inBytes = self._serial.inWaiting()
        if inBytes > 0:
            self._serial.read(inBytes)  # empty buffer and discard
        if self._byte_diff_mode:
            self._rx_buffer = None
        else:
            self._rx_buffer = ''

    def flushInput(self):
        self._serial.flushInput()

    def flushOutput(self):
        self._serial.flush()

    def write(self, bytestring):
        if type(bytestring) != bytes:
            bytestring = bytestring.encode('utf-8')
        tx_count = self._serial.write(bytestring)
        self._serial.flush()
        return tx_count

    def read(self):
        returned = self._serial.read(self._serial.inWaiting())
        if PY3:
            returned = returned.decode('utf-8')
        return returned
		
    def closeSerial(self):
        if self._serial:
            self._serial.close()
            self._serial = None
            return True
        return False

    def close(self):
        try:
            self.flushInput()
        except Exception:
            pass
        try:
            self.closeSerial()
        except Exception:
            pass
        self._serial_port = None

    def _createSerialEvent(self, logged_time, read_time, event_data):
        self._event_count += 1
        confidence_interval = read_time - self._last_poll_time
        elist = [0, 0, 0, Device._getNextEventID(),
                 EventConstants.SERIAL_INPUT,
                 read_time,
                 logged_time,
                 read_time,
                 confidence_interval,
                 0.0,
                 0,
                 self.port,
                 event_data
                 ]
        self._addNativeEventToBuffer(elist)
        self._resetParserState()

    def _createMultiByteSerialEvent(self, logged_time, read_time):
        self._event_count += 1
        confidence_interval = read_time - self._last_poll_time
        elist = [0, 0, 0, Device._getNextEventID(),
                 EventConstants.SERIAL_INPUT,
                 read_time,
                 logged_time,
                 read_time,
                 confidence_interval,
                 0.0,
                 0,
                 self.port,
                 self._parser_state['parsed_event']
                 ]
        self._addNativeEventToBuffer(elist)
        self._resetParserState()

    def _createByteChangeSerialEvent(
            self,
            logged_time,
            read_time,
            prev_byte,
            new_byte):
        self._event_count += 1
        confidence_interval = read_time - self._last_poll_time
        elist = [0, 0, 0, Device._getNextEventID(),
                 EventConstants.SERIAL_BYTE_CHANGE,
                 read_time,
                 logged_time,
                 read_time,
                 confidence_interval,
                 0.0,
                 0,
                 self.port,
                 ord(prev_byte),
                 ord(new_byte)
                 ]
        self._addNativeEventToBuffer(elist)

    def _poll(self):
        try:
            logged_time = getTime()
            if not self.isReportingEvents():
                self._last_poll_time = logged_time
                return False

            if self.isConnected():
                if self._custom_parser:
                    parser_state = self._parser_state
                    newrx = self.read()
                    read_time = getTime()
                    if newrx:
                        try:
                            serial_events = self._custom_parser(
                                read_time, newrx, parser_state, **self._custom_parser_kwargs)

                            for evt in serial_events:
                                if isinstance(evt, dict):
                                    evt_time = evt.get('time', read_time)
                                    evt_data = evt.get(
                                        'data', 'NO DATA FIELD IN EVENT DICT.')
                                    self._createSerialEvent(
                                        logged_time, evt_time, evt_data)
                                else:
                                    print2err(
                                        "ioHub Serial Device Error: Events returned from custom parser must be dict's. Skipping: ",
                                        str(evt))

                        except Exception:
                            print2err(
                                'ioHub Serial Device Error: Exception during parsing function call.')
                            import traceback
                            import sys
                            traceback.print_exc(file=sys.stderr)
                            print2err('---')

                elif self._byte_diff_mode:
                    rx = self.read()
                    read_time = getTime()
                    for c in rx:
                        if self._rx_buffer is not None and c != self._rx_buffer:
                            self._createByteChangeSerialEvent(logged_time,
                                                              read_time,
                                                              self._rx_buffer,
                                                              c)
                        self._rx_buffer = c
                else:
                    parser_state = self._parser_state
                    rx_buffer = self._rx_buffer + self.read()
                    read_time = getTime()
                    prefix = self._parser_config['prefix']
                    delimiter = self._parser_config['delimiter']

                    if parser_state['prefix_found'] is False:
                        if prefix and rx_buffer and len(
                                rx_buffer) >= len(prefix):
                            pindex = rx_buffer.find(prefix)
                            if pindex >= 0:
                                rx_buffer = rx_buffer[pindex + len(prefix):]
                                parser_state['prefix_found'] = True

                    if parser_state['delimiter_found'] is False:
                        if delimiter and self._rx_buffer and len(
                                rx_buffer) >= len(delimiter):
                            dindex = rx_buffer.find(delimiter)
                            if dindex >= 0:
                                parser_state['delimiter_found'] = True
                                sindex = dindex
                                eindex = dindex + len(delimiter)
                                parser_state[
                                    'parsed_event'] += rx_buffer[:sindex]
                                if len(rx_buffer) > eindex:
                                    rx_buffer = rx_buffer[eindex:]
                                else:
                                    rx_buffer = ''
                                self._rx_buffer = rx_buffer
                                self._createMultiByteSerialEvent(
                                    logged_time, read_time)
                                return True

                    if parser_state['bytes_needed'] > 0 and rx_buffer:
                        rxlen = len(rx_buffer)
                        #takebytes = rxlen - parser_state['bytes_needed']
                        if rxlen > parser_state['bytes_needed']:
                            parser_state[
                                'parsed_event'] += rx_buffer[:parser_state['bytes_needed']]
                            parser_state['bytes_needed'] = 0
                            rx_buffer = rx_buffer[
                                parser_state['bytes_needed']:]
                        else:
                            parser_state['parsed_event'] += rx_buffer
                            parser_state['bytes_needed'] -= rxlen
                            rx_buffer = ''

                        if parser_state['bytes_needed'] == 0:
                            self._rx_buffer = rx_buffer
                            self._createMultiByteSerialEvent(
                                logged_time, read_time)
                            return True

                    self._rx_buffer = rx_buffer
            else:
                read_time = logged_time
            self._last_poll_time = read_time
            return True
        except Exception as e:
            print2err('--------------------------------')
            print2err('ERROR in Serial._poll: ', e)
            printExceptionDetailsToStdErr()
            print2err('---------------------')

    def _close(self):
        self.setConnectionState(False)
        Device._close(self)


class Pstbox(Serial):
    """Provides convenient access to the PST Serial Response Box."""
    EVENT_CLASS_NAMES = ['PstboxButtonEvent', ]
    DEVICE_TYPE_ID = DeviceConstants.PSTBOX
    DEVICE_TYPE_STRING = 'PSTBOX'
    # Only add new attributes for the subclass, the device metaclass
    # pulls them together.
    _serial_slots = [
        '_nlamps', '_lamp_state',
        '_streaming_state', '_button_query_state', '_state',
        '_button_bytes', '_nbuttons'
    ]
    __slots__ = [e for e in _serial_slots]

    def __init__(self, *args, **kwargs):
        Serial.__init__(self, *args, **kwargs)

        self._nbuttons = 7
        # Buttons 0--4, from left to right:
        # [1, 2, 4, 8, 16]
        self._button_bytes = 2**N.arange(self._nbuttons, dtype='uint8')

        self._nlamps = 5
        self._lamp_state = N.repeat(False, self._nlamps)
        self._streaming_state = True
        self._button_query_state = True

        update_lamp_state = True
        self._state = N.r_[
            self._lamp_state, self._button_query_state, update_lamp_state,
            self._streaming_state
        ]

        self._update_state()

    def _update_state(self):
        update_lamp_state = True

        # `state` is an array of bools.
        state = N.r_[
            self._lamp_state, self._button_query_state, update_lamp_state,
            self._streaming_state
        ]

        # Convert the new state into a bitmask, collapse it into a
        # single byte and send it to the response box.
        state_bits = (2**N.arange(8))[state]
        self.write(struct.pack("B",(N.sum(state_bits))))

        # Set the `update lamp` bit to LOW again.
        state[6] = False
        self._state = state

    def getState(self):
        """Return the current state of the response box.

        Returns
        -------
        array
            An array of 8 boolean values will be returned, corresponding to
            the following properties:
            - state of lamp 0 (on/off)
            - state of lamp 1 (on/off)
            - state of lamp 2 (on/off)
            - state of lamp 3 (on/off)
            - state of lamp 4 (on/off)
            - button query state (on/off)
            - update lamp state (yes/no)
            - streaming state (on/off)

        See Also
        --------
        setState

        """
        return self._state

    def getLampState(self):
        """Return the current state of the lamps.

        Returns
        -------
        array
            An array of 5 boolean values will be returned, corresponding to
            the following properties:
            - state of lamp 0 (on/off)
            - state of lamp 1 (on/off)
            - state of lamp 2 (on/off)
            - state of lamp 3 (on/off)
            - state of lamp 4 (on/off)

        See Also
        --------
        setLampState

        """
        return self._lamp_state

    def setLampState(self, state):
        """Change the state of the lamps (on/off).

        Parameters
        ----------
        state : array_like
            The requested lamp states, from left to right. `0` or `False`
            means off, and `1` or `True` means on. This method expects
            an array of 5 boolean values, each corresponding to the
            following properties:
            - state of lamp 0 (on/off)
            - state of lamp 1 (on/off)
            - state of lamp 2 (on/off)
            - state of lamp 3 (on/off)
            - state of lamp 4 (on/off)

        See Also
        --------
        getLampState

        """
        if len(state) != self._nlamps:
            raise ValueError('Please specify a number of states that is '
                             'equal to the number of lamps on the '
                             'response box.')

        state = N.array(state).astype('bool')
        self._lamp_state = state
        self._update_state()

    def getStreamingState(self):
        """Get the current streaming state.

        Returns
        -------
        bool
            `True` if the box is streaming data, and `False` otherwise.

        See Also
        --------
        setStreamingState

        """
        return self._streaming_state

    def setStreamingState(self, state):
        """Switch on or off streaming mode.

        If streaming mode is disabled, the box will not send anything
        to the computer.

        Parameters
        ----------
        state : bool
            ``True`` will enable streaming, ``False`` will disable it.

        See Also
        --------
        getStreamingState

        """
        self._streaming_state = bool(state)
        self._update_state()

    def getButtonQueryState(self):
        """Get the current button query state.

        If the box is not querying buttons, no button presses will
        be reported.

        Returns
        -------
        bool
            `True` if the box is streaming data, and `False` otherwise.

        See Also
        --------
        getButtonQueryState

        """
        return self._button_query_state

    def setButtonQueryState(self, state):
        """Switch on or off button querying.

        Parameters
        ----------
        state : bool
            ``True`` will enable querying, ``False`` will disable it.

        See Also
        --------
        getButtonQueryState

        """
        self._button_query_state = bool(state)
        self._update_state()

    def _createByteChangeSerialEvent(self, logged_time, read_time,
                                     prev_byte, new_byte):
        self._event_count += 1
        confidence_interval = read_time - self._last_poll_time

        prev_byte = ord(prev_byte)
        new_byte = ord(new_byte)

        try:
            if new_byte != 0:  # Button was pressed
                button = N.where(self._button_bytes == new_byte)[0][0]
                button_event = 'press'
            else:  # Button was released
                button = N.where(self._button_bytes == prev_byte)[0][0]
                button_event = 'release'
        except Exception:
            # Handles when data rx does not match either N.where within the try
            return
        events = [
            0, 0, 0, Device._getNextEventID(),
            EventConstants.PSTBOX_BUTTON,
            read_time,
            logged_time,
            read_time,
            confidence_interval,
            0.0,
            0,
            self.port,
            button,
            button_event
        ]
        self._addNativeEventToBuffer(events)


class SerialInputEvent(DeviceEvent):
    _newDataTypes = [
        ('port', '|S32'),
        ('data', '|S256')
    ]
    PARENT_DEVICE = Serial
    EVENT_TYPE_ID = EventConstants.SERIAL_INPUT
    EVENT_TYPE_STRING = 'SERIAL_INPUT'
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING
    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        DeviceEvent.__init__(self, *args, **kwargs)


class SerialByteChangeEvent(DeviceEvent):
    _newDataTypes = [
        ('port', '|S32'),
        ('prev_byte', N.uint8),
        ('current_byte', N.uint8)
    ]
    PARENT_DEVICE = Serial
    EVENT_TYPE_ID = EventConstants.SERIAL_BYTE_CHANGE
    EVENT_TYPE_STRING = 'SERIAL_BYTE_CHANGE'
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING
    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        DeviceEvent.__init__(self, *args, **kwargs)


class PstboxButtonEvent(DeviceEvent):
    # Add new fields for PstboxButtonEvent
    _newDataTypes = [
        ('port', '|S32'),  # could be needed to identify events
                              # from >1 connected button box; if that is
                              # ever supported.
        ('button', N.uint8),
        ('button_event', '|S7')
    ]

    PARENT_DEVICE = Pstbox
    EVENT_TYPE_ID = EventConstants.PSTBOX_BUTTON
    EVENT_TYPE_STRING = 'PSTBOX_BUTTON'
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        DeviceEvent.__init__(self, *args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import division

from builtins import object
import sys
from .. import _ispkg

_getTime = None
MonotonicClock = None
monotonicClock = None

if _ispkg is False:
    import psychopy
    MonotonicClock = psychopy.clock.MonotonicClock
    monotonicClock = psychopy.clock.monotonicClock
    _getTime = monotonicClock.getTime
else:
    if sys.platform == 'win32':
        from ctypes import byref, c_int64, windll
        _fcounter = c_int64()
        _qpfreq = c_int64()
        windll.Kernel32.QueryPerformanceFrequency(byref(_qpfreq))
        _qpfreq = float(_qpfreq.value)
        _winQPC = windll.Kernel32.QueryPerformanceCounter

        def _getTime():
            _winQPC(byref(_fcounter))
            return _fcounter.value / _qpfreq
    elif sys.platform.startswith('linux'):
        from ctypes import byref, c_long, CDLL, Structure
        libc = CDLL('libc.so.6')
        CLOCK_MONOTONIC = 1

        class Timespec(Structure):
            _fields_ = [('tv_sec', c_long),
                        ('tv_nsec', c_long)]
        _ctime = Timespec()

        def _getTime():
            libc.clock_gettime(CLOCK_MONOTONIC, byref(_ctime))
            return _ctime.tv_sec + _ctime.tv_nsec / 1000000000.0
    elif sys.platform == "darwin":
        # Monotonic getTime with absolute origin. Suggested by @aforren1, and
        # copied from github.com/aforren1/toon/blob/master/toon/input/mac_clock.py 
        import ctypes
        _libc = ctypes.CDLL('/usr/lib/libc.dylib', use_errno=True)
        
        class mach_timebase_info_data_t(ctypes.Structure):
            _fields_ = (('numer', ctypes.c_uint32), ('denom', ctypes.c_uint32))
        
        _mach_absolute_time = _libc.mach_absolute_time
        _mach_absolute_time.restype = ctypes.c_uint64
        
        _timebase = mach_timebase_info_data_t()
        _libc.mach_timebase_info(ctypes.byref(_timebase))
        _ticks_per_second = _timebase.numer / _timebase.denom * 1.0e9
        
        def getTime():
            return _mach_absolute_time() / _ticks_per_second

    class MonotonicClock(object):
        """A convenient class to keep track of time in your experiments using a
        sub-millisecond timer.
        """

        def __init__(self, start_time=None):
            super(MonotonicClock, self).__init__()
            if start_time is None:
                # this is sub-millisec timer in python
                self._timeAtLastReset = _getTime()
            else:
                self._timeAtLastReset = start_time

        def getTime(self):
            """Returns the current time on this clock in secs (sub-ms precision)
            """
            return _getTime() - self._timeAtLastReset

        def getLastResetTime(self):
            """Returns the current offset being applied to the high resolution
            timebase used by Clock."""
            return self._timeAtLastReset

    monotonicClock = MonotonicClock()


def getTime():
    return monotonicClock.getTime()



# -*- coding: utf-8 -*-
# Copyright (c) 2011 The Chromium Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""Collection of functions and classes to fix various encoding problems on
multiple platforms with python."""
from __future__ import print_function

from builtins import str
from builtins import range
from builtins import object
import codecs
import locale
import os
import sys


# Prevents initializing multiple times.
_SYS_ARGV_PROCESSED = False


def complain(message):
    """If any exception occurs in this file, we'll probably try to print it on
    stderr, which makes for frustrating debugging if stderr is directed to our
    wrapper.

    So be paranoid about catching errors and reporting them to
    sys.__stderr__, so that the user has a higher chance to see them.

    """
    print((
        isinstance(message, str) and message or repr(message)), file=sys.__stderr__)


def fix_default_encoding():
    """Forces utf8 solidly on all platforms.

    By default python execution environment is lazy and defaults to ascii
    encoding.

    http://uucode.com/blog/2007/03/23/shut-up-you-dummy-7-bit-python/

    """
    if sys.getdefaultencoding() == 'utf-8':
        return False

    # setdefaultencoding is removed by site.py to prevent change of the
    # default encoding.  Let's import "vanilla" sys and invoke it
    # we need to remove current sys from known, so __import__ is in effect
    sys_modules = sys.modules
    our_sys = sys_modules.pop('sys')
    try:
        vanilla_sys = __import__('sys')
        # Module 'sys' has no 'setdefaultencoding' member
        # pylint: disable=E1101
        vanilla_sys.setdefaultencoding('utf-8')
        assert(sys.getdefaultencoding() == 'utf-8')
    finally:
        sys_modules['sys'] = our_sys

    for attr in dir(locale):
        if attr[0:3] != 'LC_':
            continue
        aref = getattr(locale, attr)
        try:
            locale.setlocale(aref, '')
        except locale.Error:
            continue
        try:
            lang = locale.getlocale(aref)[0]
        except (TypeError, ValueError):
            continue
        if lang:
            try:
                locale.setlocale(aref, (lang, 'UTF-8'))
            except locale.Error:
                os.environ[attr] = lang + '.UTF-8'
    try:
        locale.setlocale(locale.LC_ALL, '')
    except locale.Error:
        pass
    return True


###############################
# Windows specific


def fix_win_sys_argv(encoding):
    """Converts sys.argv to 'encoding' encoded string.

    utf-8 is recommended.

    Works around <http://bugs.python.org/issue2128>.

    """
    global _SYS_ARGV_PROCESSED
    if _SYS_ARGV_PROCESSED:
        return False

    # These types are available on linux but not Mac.
    # pylint: disable=E0611,F0401
    from ctypes import byref, c_int, POINTER, windll, WINFUNCTYPE
    from ctypes.wintypes import LPCWSTR, LPWSTR

    # <http://msdn.microsoft.com/en-us/library/ms683156.aspx>
    GetCommandLineW = WINFUNCTYPE(LPWSTR)(('GetCommandLineW', windll.kernel32))
    # <http://msdn.microsoft.com/en-us/library/bb776391.aspx>
    CommandLineToArgvW = WINFUNCTYPE(POINTER(LPWSTR), LPCWSTR, POINTER(c_int))(
        ('CommandLineToArgvW', windll.shell32))

    argc = c_int(0)
    argv_unicode = CommandLineToArgvW(GetCommandLineW(), byref(argc))
    argv = [
        argv_unicode[i].encode(encoding, 'replace')
        for i in range(0, argc.value)]

    if not hasattr(sys, 'frozen'):
        # If this is an executable produced by py2exe or bbfreeze, then it
        # will have been invoked directly. Otherwise, unicode_argv[0] is the
        # Python interpreter, so skip that.
        argv = argv[1:]

        # Also skip option arguments to the Python interpreter.
        while len(argv) > 0:
            arg = argv[0]
            if not arg.startswith(u'-') or arg == u'-':
                break
            argv = argv[1:]
            if arg == u'-m':
                # sys.argv[0] should really be the absolute path of the
                # module source, but never mind.
                break
            if arg == u'-c':
                argv[0] = u'-c'
                break
    sys.argv = argv
    _SYS_ARGV_PROCESSED = True
    return True


def fix_win_codec():
    """Works around <http://bugs.python.org/issue6058>."""
    # <http://msdn.microsoft.com/en-us/library/dd317756.aspx>
    try:
        codecs.lookup('cp65001')
        return False
    except LookupError:
        codecs.register(
            lambda name: name == 'cp65001' and codecs.lookup('utf-8-sig') or None)
        return True


class WinUnicodeOutputBase(object):
    """Base class to adapt sys.stdout or sys.stderr to behave correctly on
    Windows.

    Setting encoding to utf-8 is recommended.

    """

    def __init__(self, fileno, name, encoding):
        # Corresponding file handle.
        self._fileno = fileno
        self.encoding = encoding
        self.name = name

        self.closed = False
        self.softspace = False
        self.mode = 'w'

    @staticmethod
    def isatty():
        return False

    def close(self):
        # Don't really close the handle, that would only cause problems.
        self.closed = True

    def fileno(self):
        return self._fileno

    def flush(self):
        raise NotImplementedError()

    def write(self, text):
        raise NotImplementedError()

    def writelines(self, lines):
        try:
            for line in lines:
                self.write(line)
        except Exception as e:
            complain('%s.writelines: %r' % (self.name, e))
            raise


class WinUnicodeConsoleOutput(WinUnicodeOutputBase):
    """Output adapter to a Windows Console.

    Understands how to use the win32 console API.

    """

    def __init__(self, console_handle, fileno, stream_name, encoding):
        super(WinUnicodeConsoleOutput, self).__init__(
            fileno, '<Unicode console %s>' % stream_name, encoding)
        # Handle to use for WriteConsoleW
        self._console_handle = console_handle

        # Loads the necessary function.
        # These types are available on linux but not Mac.
        # pylint: disable=E0611,F0401
        from ctypes import byref, GetLastError, POINTER, windll, WINFUNCTYPE
        from ctypes.wintypes import BOOL, DWORD, HANDLE, LPWSTR
        from ctypes.wintypes import LPVOID  # pylint: disable=E0611

        self._DWORD = DWORD
        self._byref = byref

        # <http://msdn.microsoft.com/en-us/library/ms687401.aspx>
        self._WriteConsoleW = WINFUNCTYPE(
            BOOL, HANDLE, LPWSTR, DWORD, POINTER(DWORD), LPVOID)(
                ('WriteConsoleW', windll.kernel32))
        self._GetLastError = GetLastError

    def flush(self):
        # No need to flush the console since it's immediate.
        pass

    def write(self, text):
        try:
            if not isinstance(text, str):
                # Convert to unicode.
                text = str(text).decode(self.encoding, 'replace')
            remaining = len(text)
            while remaining > 0:
                n = self._DWORD(0)
                # There is a shorter-than-documented limitation on the length of the
                # string passed to WriteConsoleW. See
                # <http://tahoe-lafs.org/trac/tahoe-lafs/ticket/1232>.
                retval = self._WriteConsoleW(
                    self._console_handle, text,
                    min(remaining, 10000),
                    self._byref(n), None)
                if retval == 0 or n.value == 0:
                    raise IOError(
                        'WriteConsoleW returned %r, n.value = %r, last error = %r' %
                        (retval, n.value, self._GetLastError()))
                remaining -= n.value
                if not remaining:
                    break
                text = text[n.value:]
        except Exception as e:
            complain('%s.write: %r' % (self.name, e))
            raise


class WinUnicodeOutput(WinUnicodeOutputBase):
    """Output adaptor to a file output on Windows.

    If the standard FileWrite function is used, it will be encoded in
    the current code page. WriteConsoleW() permits writing any
    character.

    """

    def __init__(self, stream, fileno, encoding):
        super(WinUnicodeOutput, self).__init__(
            fileno, '<Unicode redirected %s>' % stream.name, encoding)
        # Output stream
        self._stream = stream

        # Flush right now.
        self.flush()

    def flush(self):
        try:
            self._stream.flush()
        except Exception as e:
            complain('%s.flush: %r from %r' % (self.name, e, self._stream))
            raise

    def write(self, text):
        try:
            if isinstance(text, str):
                # Replace characters that cannot be printed instead of failing.
                text = text.encode(self.encoding, 'replace')
            self._stream.write(text)
        except Exception as e:
            complain('%s.write: %r' % (self.name, e))
            raise


def win_handle_is_a_console(handle):
    """Returns True if a Windows file handle is a handle to a console."""
    # These types are available on linux but not Mac.
    # pylint: disable=E0611,F0401
    from ctypes import byref, POINTER, windll, WINFUNCTYPE
    from ctypes.wintypes import BOOL, DWORD, HANDLE

    FILE_TYPE_CHAR = 0x0002
    FILE_TYPE_REMOTE = 0x8000
    INVALID_HANDLE_VALUE = DWORD(-1).value

    # <http://msdn.microsoft.com/en-us/library/ms683167.aspx>
    GetConsoleMode = WINFUNCTYPE(BOOL, HANDLE, POINTER(DWORD))(
        ('GetConsoleMode', windll.kernel32))
    # <http://msdn.microsoft.com/en-us/library/aa364960.aspx>
    GetFileType = WINFUNCTYPE(DWORD, DWORD)(('GetFileType', windll.kernel32))

    # GetStdHandle returns INVALID_HANDLE_VALUE, NULL, or a valid handle.
    if handle == INVALID_HANDLE_VALUE or handle is None:
        return False
    return (
        (GetFileType(handle) & ~FILE_TYPE_REMOTE) == FILE_TYPE_CHAR and
        GetConsoleMode(handle, byref(DWORD())))


def win_get_unicode_stream(stream, excepted_fileno, output_handle, encoding):
    """Returns a unicode-compatible stream.

    This function will return a direct-Console writing object only if:
    - the file number is the expected console file number
    - the handle the expected file handle
    - the 'real' handle is in fact a handle to a console.
    """
    old_fileno = getattr(stream, 'fileno', lambda: None)()
    if old_fileno == excepted_fileno:
        # These types are available on linux but not Mac.
        # pylint: disable=E0611,F0401
        from ctypes import windll, WINFUNCTYPE
        from ctypes.wintypes import DWORD, HANDLE

        # <http://msdn.microsoft.com/en-us/library/ms683231.aspx>
        GetStdHandle = WINFUNCTYPE(
            HANDLE, DWORD)(
            ('GetStdHandle', windll.kernel32))

        real_output_handle = GetStdHandle(DWORD(output_handle))
        if win_handle_is_a_console(real_output_handle):
            # It's a console.
            return WinUnicodeConsoleOutput(
                real_output_handle, old_fileno, stream.name, encoding)

    # It's something else. Create an auto-encoding stream.
    return WinUnicodeOutput(stream, old_fileno, encoding)


def fix_win_console(encoding):
    """Makes Unicode console output work independently of the current code
    page.

    This also fixes <http://bugs.python.org/issue1602>.
    Credit to Michael Kaplan
    <http://blogs.msdn.com/b/michkap/archive/2010/04/07/9989346.aspx> and
    TZOmegaTZIOY
    <http://stackoverflow.com/questions/878972/windows-cmd-encoding-change-causes-python-crash/1432462#1432462>.

    """
    if (isinstance(sys.stdout, WinUnicodeOutputBase) or
            isinstance(sys.stderr, WinUnicodeOutputBase)):
        return False

    try:
        # SetConsoleCP and SetConsoleOutputCP could be used to change the code page
        # but it's not really useful since the code here is using WriteConsoleW().
        # Also, changing the code page is 'permanent' to the console and needs to be
        # reverted manually.
        # In practice one needs to set the console font to a TTF font to be able to
        # see all the characters but it failed for me in practice. In any case, it
        # won't throw any exception when printing, which is the important part.
        # -11 and -12 are defined in stdio.h
        sys.stdout = win_get_unicode_stream(sys.stdout, 1, -11, encoding)
        sys.stderr = win_get_unicode_stream(sys.stderr, 2, -12, encoding)
        # TODO(maruel): Do sys.stdin with ReadConsoleW(). Albeit the limitation is
        # "It doesn't appear to be possible to read Unicode characters in UTF-8
        # mode" and this appears to be a limitation of cmd.exe.
    except Exception as e:
        complain('exception %r while fixing up sys.stdout and sys.stderr' % e)
    return True


def fix_encoding():
    """Fixes various encoding problems on all platforms.

    Should be called at the very beginning of the process.

    """
    ret = True
    if sys.platform == 'win32':
        ret &= fix_win_codec()

    ret &= fix_default_encoding()

    if sys.platform == 'win32' and sys.version_info[0] < 3:
        encoding = sys.getdefaultencoding()
        ret &= fix_win_sys_argv(encoding)
        ret &= fix_win_console(encoding)
    return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import wx
import requests

from psychopy import logging
from .. import dialogs
from .functions import logInPavlovia
from psychopy.app.pavlovia_ui.project import syncProject
from .search import SearchFrame
from .project import ProjectEditor
from psychopy.localization import _translate
from psychopy.projects import pavlovia


class PavloviaMenu(wx.Menu):
    app = None
    appData = None
    currentUser = None
    knownUsers = None
    searchDlg = None

    def __init__(self, parent):
        wx.Menu.__init__(self)
        self.parent = parent  # type: BuilderFrame
        PavloviaMenu.app = parent.app
        keys = self.app.keys
        # from prefs fetch info about prev usernames and projects
        PavloviaMenu.appData = self.app.prefs.appData['projects']

        # item = self.Append(wx.ID_ANY, _translate("Tell me more..."))
        # parent.Bind(wx.EVT_MENU, self.onAbout, id=item.GetId())

        PavloviaMenu.knownUsers = pavlovia.knownUsers

        # sub-menu for usernames and login
        self.userMenu = wx.Menu()
        # if a user was previously logged in then set them as current
        lastPavUser = PavloviaMenu.appData['pavloviaUser']
        if pavlovia.knownUsers and (lastPavUser not in pavlovia.knownUsers):
            lastPavUser = None
        # if lastPavUser and not PavloviaMenu.currentUser:
        #     self.setUser(PavloviaMenu.appData['pavloviaUser'])
        for name in self.knownUsers:
            self.addToSubMenu(name, self.userMenu, self.onSetUser)
        self.userMenu.AppendSeparator()
        self.loginBtn = self.userMenu.Append(wx.ID_ANY,
                                    _translate("Log in to Pavlovia...\t{}")
                                    .format(keys['pavlovia_logIn']))
        parent.Bind(wx.EVT_MENU, self.onLogInPavlovia, id=self.loginBtn.GetId())
        self.AppendSubMenu(self.userMenu, _translate("User"))

        # search
        self.searchBtn = self.Append(wx.ID_ANY,
                           _translate("Search Pavlovia\t{}")
                           .format(keys['projectsFind']))
        parent.Bind(wx.EVT_MENU, self.onSearch, id=self.searchBtn.GetId())

        # new
        self.newBtn = self.Append(wx.ID_ANY,
                           _translate("New...\t{}").format(keys['projectsNew']))
        parent.Bind(wx.EVT_MENU, self.onNew, id=self.newBtn.GetId())

        self.syncBtn = self.Append(wx.ID_ANY,
                           _translate("Sync\t{}").format(keys['projectsSync']))
        parent.Bind(wx.EVT_MENU, self.onSync, id=self.syncBtn.GetId())

    def addToSubMenu(self, name, menu, function):
        item = menu.Append(wx.ID_ANY, name)
        self.parent.Bind(wx.EVT_MENU, function, id=item.GetId())

    def onAbout(self, event):
        wx.GetApp().followLink(event)

    def onSetUser(self, event):
        user = self.userMenu.GetLabelText(event.GetId())
        self.setUser(user)

    def setUser(self, user=None):

        if user is None and PavloviaMenu.appData['pavloviaUser']:
            user = PavloviaMenu.appData['pavloviaUser']

        if user in [PavloviaMenu.currentUser, None]:
            return  # nothing to do here. Move along please.

        PavloviaMenu.currentUser = user
        PavloviaMenu.appData['pavloviaUser'] = user
        if user in pavlovia.knownUsers:
            token = pavlovia.knownUsers[user]['token']
            try:
                pavlovia.getCurrentSession().setToken(token)
            except requests.exceptions.ConnectionError:
                logging.warning("Tried to log in to Pavlovia but no network "
                                "connection")
                return
        else:
            if hasattr(self, 'onLogInPavlovia'):
                self.onLogInPavlovia()

        if PavloviaMenu.searchDlg:
            PavloviaMenu.searchDlg.updateUserProjs()

    def onSync(self, event):
        retVal = syncProject(parent=self.parent, project=self.parent.project)
        if hasattr(self.parent, 'gitFeedback'):
            self.parent.gitFeedback(retVal)

    def onSearch(self, event):
        PavloviaMenu.searchDlg = SearchFrame(app=self.parent.app)
        PavloviaMenu.searchDlg.Show()

    def onLogInPavlovia(self, event=None):
        logInPavlovia(parent=self.parent)

    def onNew(self, event):
        """Create a new project
        """
        if pavlovia.getCurrentSession().user.username:
            projEditor = ProjectEditor()
            if projEditor.ShowModal() == wx.ID_OK:
                self.parent.project = projEditor.project
                # do a first sync as well
                retVal = syncProject(parent=self.parent, project=projEditor.project)
                self.parent.gitFeedback(retVal)
        else:
            infoDlg = dialogs.MessageDialog(parent=None, type='Info',
                                            message=_translate(
                                                    "You need to log in"
                                                    " to create a project"))
            infoDlg.Show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2020 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from past.builtins import basestring

from .project import DetailsPanel
from .functions import logInPavlovia

from ._base import BaseFrame
from psychopy.localization import _translate
from psychopy.projects import pavlovia

import copy
import wx
from pkg_resources import parse_version
from wx.lib import scrolledpanel as scrlpanel
import wx.lib.mixins.listctrl as listmixin

import requests

starChar = u"\u2B50"
forkChar = u"\u2442"


class SearchFrame(wx.Dialog):

    def __init__(self, app, parent=None, style=None,
                 pos=wx.DefaultPosition):
        if style is None:
            style = (wx.DEFAULT_DIALOG_STYLE | wx.CENTER |
                     wx.TAB_TRAVERSAL | wx.RESIZE_BORDER)
        title = _translate("Search for projects online")
        self.frameType = 'ProjectSearch'
        wx.Dialog.__init__(self, parent, -1, title=title, style=style,
                           size=(800, 500), pos=pos)

        self.app = app
        self.project = None
        self.parent = parent
        # info about last search (NB None means no search but [] or '' means empty)
        self.lastSearchStr = None
        self.lastSearchOwn = None
        self.lastSearchGp = None
        self.lastSearchPub = None

        # self.mainPanel = wx.Panel(self, wx.ID_ANY)
        self.searchLabel = wx.StaticText(self, wx.ID_ANY, _translate('Search:'))
        self.searchCtrl = wx.TextCtrl(self, wx.ID_ANY)
        self.searchCtrl.Bind(wx.EVT_BUTTON, self.onSearch)
        self.searchBtn = wx.Button(self, wx.ID_ANY, _translate("Search"))
        self.searchBtn.Bind(wx.EVT_BUTTON, self.onSearch)
        self.searchBtn.SetDefault()

        self.searchInclPublic = wx.CheckBox(self, wx.ID_ANY,
                                          label="Public")
        self.searchInclPublic.Bind(wx.EVT_CHECKBOX, self.onSearch)
        self.searchInclPublic.SetValue(True)

        self.searchInclGroup = wx.CheckBox(self, wx.ID_ANY,
                                          label="My groups")
        self.searchInclGroup.Bind(wx.EVT_CHECKBOX, self.onSearch)
        self.searchInclGroup.SetValue(True)

        self.searchBuilderOnly = wx.CheckBox(self, wx.ID_ANY,
                                             label="Only Builder")
        self.searchBuilderOnly.Bind(wx.EVT_CHECKBOX, self.onSearch)
        # then the search results
        self.searchResults = ProjectListCtrl(self)

        # on the right
        self.detailsPanel = DetailsPanel(parent=self)

        # sizers layout
        self.searchBtnSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.searchBtnSizer.Add(self.searchCtrl, 1, wx.EXPAND | wx.ALL, 5)
        self.searchBtnSizer.Add(self.searchBtn, 0, wx.EXPAND | wx.ALL, 5)
        self.optionsSizer = wx.WrapSizer()
        self.optionsSizer.AddMany([self.searchInclGroup, self.searchInclPublic,
                                   self.searchBuilderOnly])

        self.leftSizer = wx.BoxSizer(wx.VERTICAL)
        self.leftSizer.Add(self.searchLabel, 0, wx.EXPAND | wx.ALL, 5)
        self.leftSizer.Add(self.optionsSizer)
        self.leftSizer.Add(self.searchBtnSizer, 0, wx.EXPAND | wx.ALL, 5)
        self.leftSizer.Add(self.searchResults, 1, wx.EXPAND | wx.ALL, 5)

        self.mainSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.mainSizer.Add(self.leftSizer, 1, wx.EXPAND | wx.ALL, 5)
        self.mainSizer.Add(self.detailsPanel, 1, wx.EXPAND | wx.ALL, 5)

        self.SetSizer(self.mainSizer)  # don't fit until search is populated
        if self.parent:
            self.CenterOnParent()
        self.Layout()
        # if projects == 'no user':
        #     msg = _translate("Log in to search your own projects")
        #     loginBtn = wx.Button(self, wx.ID_ANY, label=msg)
        #     loginBtn.Bind(wx.EVT_BUTTON, self.onLoginClick)

    def getSearchOptions(self):
        opts = {}
        opts['inclPublic'] = self.searchInclPublic.GetValue()
        opts['builderOnly'] = self.searchBuilderOnly.GetValue()
        opts['inclGroup'] = self.searchBuilderOnly.GetValue()
        return opts

    def onSearch(self, evt=None):
        opts = self.getSearchOptions()

        searchStr = self.searchCtrl.GetValue()
        newSearch = (searchStr!=self.lastSearchStr)
        self.lastSearchStr = newSearch

        session = pavlovia.getCurrentSession()
        # search own
        if newSearch:
            try:
                self.lastSearchOwn = session.gitlab.projects.list(owned=True, search=searchStr)
            except requests.exceptions.ConnectionError:
                print("Failed to establish a new connection: No internet?")
                return None

        # search my groups
        if opts['inclGroup'] and (newSearch or self.lastSearchGp is None):
            # group projects: owned=False, membership=True
            self.lastSearchGp = session.gitlab.projects.list(
                owned=False, membership=True, search=searchStr)
        elif not opts['inclGroup']:  # set to None (to indicate non-search not simply empty result)
            self.lastSearchGp = None
        elif opts['inclGroup'] and not newSearch:
            pass  # we have last search and we need it so do nothing
        else:
            print("ERROR: During Pavlovia search we found opts['inclGroup']={}, newSearch={}"
                  .format(opts['inclGroup'], newSearch))

        # search public
        if opts['inclPublic'] and (newSearch or self.lastSearchPub is None):
            self.lastSearchPub = session.gitlab.projects.list(owned=False, membership=False,
                                                              search=searchStr)
        elif not opts['inclPublic']:  # set to None (to indicate non-search not simply empty result)
            self.lastSearchPub = None
        elif opts['inclPublic'] and not newSearch:
            pass  # we have last search and we need it so do nothing
        else:
            print("ERROR: During Pavlovia search we found opts['inclPublic']={}, newSearch={}"
                  .format(opts['inclPublic'], newSearch))

        projs = copy.copy(self.lastSearchOwn)
        if opts['inclGroup']:
            projs.extend(self.lastSearchGp)
        if opts['inclPublic']:
            projs.extend(self.lastSearchPub)

        projs = getUniqueByID(projs)
        projs = [pavlovia.PavloviaProject(proj) for proj in projs if proj.id]

        self.searchResults.setContents(projs)
        self.searchResults.Update()
        self.Layout()

    def onLoginClick(self, event):
        user = logInPavlovia(parent=self.parent)

    def Show(self):
        # show the dialog then start search
        wx.Dialog.Show(self)
        wx.Yield()
        self.onSearch()  # trigger the search update


class ProjectListCtrl(wx.ListCtrl, listmixin.ListCtrlAutoWidthMixin):
    """A scrollable panel showing a list of projects. To be used within the
    Project Search dialog
    """

    def __init__(self, parent, frame=None):
        wx.ListCtrl.__init__(self, parent, wx.ID_ANY,
                             style=wx.LC_REPORT | wx.LC_SINGLE_SEL)
        listmixin.ListCtrlAutoWidthMixin.__init__(self)
        self.AlwaysShowScrollbars(True)
        self.parent = parent
        if frame is None:
            self.frame = parent
        else:
            self.frame = frame
        self.projList = []
        self.columnNames = [starChar, forkChar, 'Group', 'Name', 'Description']
        self._currentSortCol = 0
        self._currentSortRev = False

        # Give it some columns.
        # The ID col we'll customize a bit:
        for n, columnName in enumerate(self.columnNames):
            if len(columnName) < 3:  # for short names center the text
                self.InsertColumn(n, columnName, wx.LIST_FORMAT_CENTER)
            else:
                self.InsertColumn(n, columnName)
        # set the column sizes *after* adding the items
        for n, columnName in enumerate(self.columnNames):
            self.SetColumnWidth(n, wx.LIST_AUTOSIZE)

        # after creating columns we can create the sort mixin
        # listmixin.ColumnSorterMixin.__init__(self, len(columnList))
        self.SetAutoLayout(True)
        self.Bind(wx.EVT_LIST_COL_CLICK, self.onColumnClick)
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.onChangeSelection)

    def setContents(self, projects):
        self.DeleteAllItems()
        # first time around we have a list of PavloviaProjects
        if projects and isinstance(projects[0], pavlovia.PavloviaProject):
            self.projList = []
            for index, thisProj in enumerate(projects):
                if not hasattr(thisProj, 'id'):
                    continue
                data = (thisProj.star_count, thisProj.forks_count,
                        thisProj.group, thisProj.name, thisProj.description)
                proj = {}
                proj[starChar] = thisProj.star_count
                proj[forkChar] = thisProj.forks_count
                proj['Name'] = thisProj.name
                proj['Group'] = thisProj.group
                proj['Description'] = thisProj.description
                proj['id'] = thisProj.id
                self.projList.append(proj)
                self.Append(data)  # append to the wx table
        # subsequent iterations are simple dicts
        else:
            self.projList = projects
            for index, thisProj in enumerate(projects):
                data = (thisProj[starChar], thisProj[forkChar],
                        thisProj['Group'], thisProj['Name'],
                        thisProj['Description'])
                self.Append(data)  # append to the wx table
        self.resizeCols(finalOnly=False)
        self.Update()

    def resizeCols(self, finalOnly):
        # resize the columns
        for n in range(self.ColumnCount):
            if not finalOnly:
                self.SetColumnWidth(n, wx.LIST_AUTOSIZE_USEHEADER)
                if self.GetColumnWidth(n) > 100:
                    self.SetColumnWidth(n, 100)

    def onChangeSelection(self, event):
        proj = self.projList[event.GetIndex()]
        self.frame.detailsPanel.setProject(proj['id'])

    def onColumnClick(self, event=None):
        col = event.Column
        if col == self._currentSortCol:  # toggle direction
            self._currentSortRev = not(self._currentSortRev)
        self._currentSortCol = col
        projs = sortProjects(self.projList, self.columnNames[col],
                             reverse=self._currentSortRev)
        self.setContents(projs)


def sortProjects(seq, name, reverse=False):
    return sorted(seq, key=lambda k: k[name], reverse=reverse)

def getUniqueByID(seq):
    """Very fast function to remove duplicates from a list while preserving order

    Based on sort f8() by Dave Kirby
    benchmarked at https://www.peterbe.com/plog/uniqifiers-benchmark

    Requires Python>=2.7 (requires set())
    """
    # Order preserving
    seen = set()
    return [x for x in seq if x.id not in seen and not seen.add(x.id)]
